import functools
from typing import NamedTuple
import jax
import jax.lax as lax
import jax.numpy as jnp
from einops import rearrange
import chex


class Carry(NamedTuple):
    numerator: chex.Array
    denominator: chex.Array
    max_so_far: chex.Array


def efficient_attention(
        query: chex.Array,
        key: chex.Array,
        value: chex.Array,
        bias: chex.Array = None,
        deterministic: bool = True,
        dropout_rng: chex.PRNGKey = None,
        attention_drop_rate: float = 0.0,
        causal: bool = True,
        query_chunk_size: int = 1024,
        key_chunk_size: int = 1024,
        dtype: chex.ArrayDType = jnp.float32,
        policy=jax.checkpoint_policies.nothing_saveable(),
        precision=None,
        float32_logits: bool = True,
        prevent_cse: bool = True,
):
    """

    :param query: Array Shape [batch,Q Sequence length,num attention heads, head dims]
    :param key: Array Shape [batch,KV Sequence length,num KV attention heads, head dims]
    :param value: Array Shape [batch,KV Sequence length,num KV attention heads, head dims]
    :param bias: Bias To be added
    :param deterministic: bool (whenever use dropout or no)
    :param dropout_rng: RNG Dropout
    :param attention_drop_rate:
    :param causal: Is Decoder or Causal
    :param query_chunk_size: Chunk size used for query
    :param key_chunk_size: Chunk size used for key
    :param dtype: DataType
    :param policy: Gradient Checkpoint Policy
    :param precision: PrecisionLike
    :param float32_logits:
    :param prevent_cse:
    :return:
    """
    query = query / jnp.sqrt(query.shape[-1]).astype(dtype)
    if float32_logits:
        query = query.astype(jnp.float32)
        key = key.astype(jnp.float32)

    batch, q_len, num_heads, dim_per_head = query.shape
    batch, kv_len, kv_heads, dim_per_head = key.shape
    batch, kv_len, kv_heads, dim_per_head = value.shape

    num_q = q_len // query_chunk_size
    num_kv = kv_len // key_chunk_size
    query = query.reshape((batch, num_q, query_chunk_size, num_heads, dim_per_head))
    key = key.reshape((batch, num_kv, key_chunk_size, kv_heads, dim_per_head))
    value = value.reshape((batch, num_kv, key_chunk_size, kv_heads, dim_per_head))

    query = jnp.moveaxis(query, 1, 0)
    key = jnp.moveaxis(key, 1, 0)
    value = jnp.moveaxis(value, 1, 0)

    if bias is not None:
        for bias_dim, broadcast_dim in zip(bias.shape, (batch, num_heads, q_len, kv_len)):
            assert bias_dim == 1 or bias_dim == broadcast_dim
    if not deterministic and attention_drop_rate > 0.0:
        attn_dropout_rng, dropout_rng = jax.random.split(dropout_rng)
        attn_dropout = jax.random.bernoulli(attn_dropout_rng, attention_drop_rate, (batch, num_heads, q_len, kv_len))
    else:
        attn_dropout = None

    _chunk_bias_fn = functools.partial(
        _chunk_attention_bias,
        query_chunk_size, key_chunk_size, bias, deterministic,
        attn_dropout, attention_drop_rate, causal, dtype)

    def scan_attention(args):
        query_chunk, query_chunk_idx = args

        @functools.partial(jax.checkpoint, prevent_cse=prevent_cse, policy=policy)
        def scan_kv_block(carry, args):
            key_chunk, value_chunk, key_chunk_idx = args
            (numerator, denominator, prev_max_score) = carry
            attn_weights = jnp.einsum('bqhd,bkhd->bqhk', query_chunk, key_chunk, precision=precision)
            bias_chunk = _chunk_bias_fn(query_chunk_idx, key_chunk_idx)
            bias_chunk = jnp.moveaxis(bias_chunk, 1, 2)
            attn_weights = attn_weights + bias_chunk

            max_score = jnp.max(attn_weights, axis=-1, keepdims=True)
            max_score = jnp.maximum(prev_max_score, max_score)
            max_score = jax.lax.stop_gradient(max_score)
            exp_weights = jnp.exp(attn_weights - max_score)
            exp_values = jnp.einsum(
                'bqhv,bvhd->bqhd', exp_weights, value_chunk, precision=precision
            )
            correction = jnp.exp(prev_max_score - max_score)
            numerator = numerator * correction + exp_values
            denominator = denominator * correction + exp_weights.sum(axis=-1, keepdims=True)
            return Carry(numerator, denominator, max_score), None

        def skip_upper_half(carry, args):
            key_chunk, value_chunk, key_chunk_idx = args
            skip_block = jnp.array(False)
            if causal:
                skip_block = query_chunk_idx < key_chunk_idx
            return jax.lax.cond(
                skip_block,
                lambda carry, args: (carry, None),
                scan_kv_block,
                carry,
                args,
            )

        init_carry = Carry(
            jnp.zeros((batch, query_chunk_size, num_heads, dim_per_head), dtype=query.dtype),
            jnp.zeros((batch, query_chunk_size, num_heads, dim_per_head), dtype=query.dtype),
            (-jnp.inf) * jnp.ones((batch, query_chunk_size, num_heads, 1), dtype=query.dtype),
        )
        (numerator, denominator, max_score), _ = lax.scan(
            skip_upper_half, init_carry, xs=(key, value, jnp.arange(0, num_kv))
        )
        outputs = (numerator / denominator).astype(dtype)
        return outputs

    _, res = lax.scan(
        lambda _, x: ((), scan_attention(x)),
        (), xs=(query, jnp.arange(0, num_q))
    )
    res = rearrange(res, 'n b c h d -> b (n c) h d')
    return res


def _chunk_attention_bias(
        query_chunk_size: int,
        key_chunk_size: int,
        bias: chex.Array,
        deterministic: bool,
        attn_dropout: chex.Array,
        attention_drop_rate: float,
        causal: bool,
        dtype: chex.ArrayDType,
        query_chunk_idx: int,
        key_chunk_idx: int
):
    """
    The _chunk_attention_bias function is used to compute the attention bias for a single chunk of
    the query and key tensors. The function takes in the following arguments:

    :param query_chunk_size: int: Determine the size of the query chunk
    :param key_chunk_size: int: Determine the size of the key_chunk
    :param bias: chex.Array: Mask out the attention weights
    :param deterministic: bool: Determine whether to use dropout or not
    :param attn_dropout: chex.Array: Drop out attention weights
    :param attention_drop_rate: float: Determine the dropout rate for attention
    :param causal: bool: Determine if the attention is causal or not
    :param dtype: chex.ArrayDType: Specify the data type of the array
    :param query_chunk_idx: int: Select the query chunk
    :param key_chunk_idx: int: Determine the key_offset
    :return: A chunk of the attention bias
    
    """
    query_offset = query_chunk_idx * query_chunk_size
    key_offset = key_chunk_idx * key_chunk_size
    chunk_bias = jnp.zeros((1, 1, 1, 1), dtype=dtype)
    if bias is not None:
        chunk_bias = lax.dynamic_slice(
            bias,
            start_indices=(0, 0, query_offset, key_offset),
            slice_sizes=(*bias.shape[:2], min(bias.shape[-2], query_chunk_size), min(bias.shape[-1], key_chunk_size)),
        )

    if causal:
        query_idx = lax.broadcasted_iota(dtype=jnp.int32, shape=(query_chunk_size, 1), dimension=0)
        key_idx = lax.broadcasted_iota(dtype=jnp.int32, shape=(1, key_chunk_size), dimension=1)
        offset = query_offset - key_offset
        query_idx += offset
        causal_mask_value = (query_idx < key_idx) * jnp.finfo(dtype).min
        chunk_bias += causal_mask_value.reshape(1, 1, *causal_mask_value.shape)

    if not deterministic and attention_drop_rate > 0.0:
        attn_dropout_slice = lax.dynamic_slice(
            attn_dropout,
            start_indices=(0, 0, query_offset, key_offset),
            slice_sizes=(
                *attn_dropout.shape[:2],
                min(attn_dropout.shape[-2], query_chunk_size),
                min(attn_dropout.shape[-1], key_chunk_size),
            ),
        )
        chunk_bias += attn_dropout_slice * jnp.finfo(dtype).min
    return chunk_bias.astype(dtype)



# Why Do We have BITs here?

# Bits is a clone of https://github.com/google/aqt

# AQT library by google, but I needed to be able to change a lot of parts of the code and I couldn't do this on Google
# Repo for sure so I just made a copy of the part of the library that I wanted (AQT is Apache 2 Licenced Project)
# Ill more focus on doing the job for the llm checkpoint in 8 bit so there's no need to project name be AQT Accurate QTraining

from .bits import matmul_true_int8, matmul, q_matmul_int8
from .q_dot_general import make_dot_general, make_fake_quant, DotGeneralRes
from .q_flax import QuantMode, Freezer, QDotGeneral, QEinsum, config_v4
from .config import fully_quantized



import optax
from typing import NamedTuple
from jax import numpy as jnp
import jax
import chex


class OptaxScheduledWeightDecayState(NamedTuple):
    count: chex.Array


def optax_add_scheduled_weight_decay(schedule_fn, mask=None):
    """

    :param schedule_fn:
    :param mask:
    :return: Optax GradientTransformation inited
    """

    def init_fn(params):
        del params
        return OptaxScheduledWeightDecayState(count=jnp.zeros([], jnp.int32))

    def update_fn(updates, state, params):
        if params is None:
            raise ValueError('Params cannot be None for weight decay!')

        weight_decay = schedule_fn(state.count)
        updates = jax.tree_util.tree_map(
            lambda g, p: g + weight_decay * p, updates, params
        )
        return updates, OptaxScheduledWeightDecayState(
            count=optax.safe_int32_increment(state.count)
        )

    if mask is not None:
        return optax.masked(optax.GradientTransformation(init_fn, update_fn), mask)
    return optax.GradientTransformation(init_fn, update_fn)



from typing import Optional
import chex
import optax


def get_lion_with_linear_scheduler(
        steps: int,
        learning_rate_start: float = 5e-5,
        learning_rate_end: float = 1e-5,
        b1: float = 0.9,
        b2: float = 0.99,
        gradient_accumulation_steps: int = 1,
        mu_dtype: Optional[chex.ArrayDType] = None,
):
    """
    Args:
        steps: total train steps (max_steps)
        learning_rate_start: start learning rate for sure
        learning_rate_end: end learning rate for sure :\
        b1: Rate for combining the momentum and the current grad.
        b2: Decay rate for the exponentially weighted average of grads.
        mu_dtype: Optional `dtype` to be used for the momentum; if
          `None` then the `dtype is inferred from `params` and `updates`.
        gradient_accumulation_steps:gradient_accumulation_steps
    Return:
        Optimizer , Scheduler
         """
    scheduler = optax.linear_schedule(
        init_value=learning_rate_start,
        end_value=learning_rate_end,
        transition_steps=steps
    )
    tx = optax.chain(
        optax.scale_by_lion(
            b1=b1,
            b2=b2,
            mu_dtype=mu_dtype
        ),
        optax.scale_by_schedule(scheduler),
        optax.scale(-1)
    )
    if gradient_accumulation_steps > 1:
        tx = optax.MultiSteps(
            tx, gradient_accumulation_steps
        )
    return tx, scheduler


def get_lion_with_with_warmup_linear_scheduler(
        steps: int,
        b1: float = 0.9,
        b2: float = 0.99,
        gradient_accumulation_steps: int = 1,
        mu_dtype: Optional[chex.ArrayDType] = None,
        learning_rate_start: float = 5e-5,
        learning_rate_end: float = 1e-5,
        warmup_steps: int = 500
):
    """

    :param b1:
    :param b2:
    :param mu_dtype:
    :param learning_rate_start:
    :param learning_rate_end:
    :param warmup_steps:
    :param gradient_accumulation_steps:
    :param steps:
    :param gradient_accumulation_steps
    :return: Optimizer and Scheduler
    """
    scheduler_warmup = optax.linear_schedule(init_value=5e-8, end_value=learning_rate_start,
                                             transition_steps=warmup_steps)
    scheduler_decay = optax.linear_schedule(init_value=learning_rate_start, end_value=learning_rate_end,
                                            transition_steps=steps - warmup_steps)

    scheduler_combined = optax.join_schedules(schedules=[scheduler_warmup, scheduler_decay], boundaries=[warmup_steps])
    tx = optax.chain(
        optax.scale_by_lion(
            b1=b1,
            b2=b2,
            mu_dtype=mu_dtype
        ),
        optax.scale_by_schedule(scheduler_combined),
        optax.scale(-1)
    )
    if gradient_accumulation_steps > 1:
        tx = optax.MultiSteps(
            tx, gradient_accumulation_steps
        )
    return tx, scheduler_combined


def get_lion_with_cosine_scheduler(
        steps: int,
        learning_rate=5e-5,
        alpha: float = 0.0,
        exponent: float = 1.0,
        b1: float = 0.9,
        b2: float = 0.99,
        gradient_accumulation_steps: int = 1,
        mu_dtype: Optional[chex.ArrayDType] = None,
):
    """

    Args:
        learning_rate: An initial value `init_v`.
        steps: Positive integer - the number of steps for which to apply
          the decay for.
        alpha: Float. The minimum value of the multiplier used to adjust the
          learning rate.
        exponent: Float. The default decay is 0.5 * (1 + cos(pi * t/T)), where t is
          the current timestep and T is the `decay_steps`. The exponent modifies
          this to be (0.5 * (1 + cos(pi * t/T))) ** exponent. Defaults to 1.0.
        b1: Rate for combining the momentum and the current grad.
        b2: Decay rate for the exponentially weighted average of grads.
        mu_dtype: Optional `dtype` to be used for the momentum; if
          `None` then the `dtype is inferred from `params` and `updates`.
        gradient_accumulation_steps:gradient_accumulation_steps
    Return:
        Optimizer , Scheduler

    """
    try:
        scheduler = optax.cosine_decay_schedule(
            init_value=learning_rate,
            decay_steps=steps,
            alpha=alpha,
            exponent=exponent
        )
    except:
        scheduler = optax.cosine_decay_schedule(
            init_value=learning_rate,
            decay_steps=steps,
            alpha=alpha,
            # exponent=exponent
        )
    tx = optax.chain(
        optax.scale_by_lion(
            b1=b1,
            b2=b2,
            mu_dtype=mu_dtype
        ),
        optax.scale_by_schedule(scheduler),
        optax.scale(-1)
    )
    if gradient_accumulation_steps > 1:
        tx = optax.MultiSteps(
            tx, gradient_accumulation_steps
        )
    return tx, scheduler


def get_lion_with_warm_up_cosine_scheduler(
        steps: int,
        learning_rate=5e-5,
        exponent: float = 1.0,
        b1: float = 0.9,
        b2: float = 0.99,
        gradient_accumulation_steps: int = 1,
        mu_dtype: Optional[chex.ArrayDType] = None,
):
    """

    :param steps:
    :param learning_rate:
    :param exponent:
    :param b1:
    :param b2:
    :param gradient_accumulation_steps:
    :param mu_dtype:
    :return:
    """
    scheduler = optax.warmup_cosine_decay_schedule(
        init_value=0.5e-7,
        peak_value=learning_rate,
        warmup_steps=steps,
        decay_steps=steps + 1,
        end_value=learning_rate,
        exponent=exponent,
    )
    tx = optax.chain(
        optax.scale_by_lion(
            b1=b1,
            b2=b2,
            mu_dtype=mu_dtype
        ),
        optax.scale_by_schedule(scheduler),
        optax.scale(-1)
    )
    if gradient_accumulation_steps > 1:
        tx = optax.MultiSteps(
            tx, gradient_accumulation_steps
        )
    return tx, scheduler



"""
# File: __init__.py

## Purpose:
This file contains imports for various attention mechanisms used in our project.

## Imports:
1. `efficient_attention` from `efficient_attention` module: This import brings in the `efficient_attention` function,
 which implements an efficient attention mechanism for our project.

2. `dot_product_attention_multihead`, `dot_product_attention_multiquery`, `dot_product_attention_queries_per_head` from
 `flash_attention_0` module: These imports bring in different variations of dot product attention mechanisms for
 multi-head and multi-query scenarios.

3. `ring_attention`, `ring_attention_standard`, `ring_flash_attention_gpu`, `ring_flash_attention_tpu`, `blockwise_ffn`,
 `blockwise_attn` from `flash_attention` module: These imports bring in various attention mechanisms such as
 ring attention, flash attention for GPU and TPU, and blockwise feed-forward network and attention mechanisms.

4. `tpu_flash_attention` `gpu_flash_attention` are imported from jax_flash_attn_tpu/gpu

"""

from .efficient_attention import efficient_attention as efficient_attention
from .flash_attention_0 import (
    dot_product_attention_multihead as dot_product_attention_multihead,
    dot_product_attention_multiquery as dot_product_attention_multiquery,
    dot_product_attention_queries_per_head as dot_product_attention_queries_per_head
)
from .flash_attention import (
    ring_attention as ring_attention,
    ring_attention_standard as ring_attention_standard,
    ring_flash_attention_gpu as ring_flash_attention_gpu,
    ring_flash_attention_tpu as ring_flash_attention_tpu,
    blockwise_ffn as blockwise_ffn,
    blockwise_attn as blockwise_attn
)
from .jax_flash_attn_tpu import flash_attention as tpu_flash_attention, BlockSizes as BlockSizes
from .jax_flash_attn_gpu import mha as gpu_flash_attention



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantized dot_general."""

# Lingo in this file:
#
# - lhs(rhs) - left(right) hand side of a binary operation
# - ca - contraction axes
# - ba - batch axes
# - ra - remaining axes

# pylint: disable=g-explicit-bool-comparison
# pylint: disable=g-explicit-length-test

import functools
from typing import Callable, Optional, Union
from . import config
from . import no_numerics
import flax.struct
import jax
from jax import lax
import jax.numpy as jnp
import numpy as onp


@flax.struct.dataclass
class Context:
    key: Optional[jax.Array]
    train_step: Optional[int]


def _context_split(context: Context) -> tuple[Context, Context]:
    def mk_ctx(key):
        return Context(key=key, train_step=context.train_step)

    if context.key is not None:
        key1, key2 = jax.random.split(context.key)
        return mk_ctx(key1), mk_ctx(key2)
    return mk_ctx(None), mk_ctx(None)


def _scale_quant(x, *, cfg, ca, context):
    """The core quantizing function."""
    msg = (
        'use_fake_quant mode is used in tests and it is exactly equal when'
        ' po2_scale == True; Did you forget to set it?'
    )
    assert (not cfg.use_fake_quant) or cfg.po2_scale, msg

    # TODO(lew): We should cast earlier. xhs_q should be in cfg.xhs.dtype
    # TODO(lew): After we implement optimization to not double-quantize,
    #   what would happen if we pass fq value (xhs_q2) in residual?

    if isinstance(cfg.numerics, no_numerics.NoNumerics):
        return x, None, None
    shared_axes = cfg.calib_shared_axes or ca
    bound = cfg.calibration.get_bound(x, shared_axes)
    abs_max_mapped_to = cfg.numerics.abs_val_mapped_to()
    scale = abs_max_mapped_to / bound

    if cfg.po2_scale:
        # With floor the biggest value (we are using jnp.max) is in the range of
        # clipping and therefore have a correct gradinet.
        scale = 2 ** jnp.floor(jnp.log2(scale))
    if cfg.scale_stop_grad:
        # TODO(lew): Does not matter in DG, because we are using custom gradient.
        #   We should take that into account somehow.
        scale = lax.stop_gradient(scale)

    x_s = _maybe_mul(x, scale)

    quant = jax.custom_vjp(cfg.numerics.fwd)
    quant.defvjp(cfg.numerics.vjp_fwd, cfg.numerics.vjp_bwd)
    quant = functools.partial(quant, context=context)

    x_q, quant_grad = jax.vjp(quant, x_s)
    # We are passing quant_grad (and not more) ot the backward pass.
    # That is equivalent to having:
    # scale = stop_gradient(scale)
    #
    # This is not the only possible choice and we intend to allow experimentation.
    # However for today we hardcoded this choice.
    #
    # In order to achevie no-stop-gradiend solution, we should take vjp
    # of a larger piece of code like the whole _scale_quant.
    #
    # TODO(lew): Implement configuration of stop-gradient.
    inv_scale = _maybe_inv(scale)

    return x_q, inv_scale, quant_grad


def make_fake_quant(cfg: config.Tensor, ca=None):
    def fake_quant(x, context):
        x_q, inv_scale, _ = _scale_quant(x, cfg=cfg, ca=ca, context=context)
        return _maybe_mul(x_q, inv_scale)

    return fake_quant


@flax.struct.dataclass
# It is used only when use_fwd_quant = True
class QTensor:
    qvalue: jnp.ndarray
    qvalue_scale_t: jnp.ndarray


@flax.struct.dataclass
class MultiTensor:
    x: jnp.ndarray
    qx: Optional[QTensor]


@flax.struct.dataclass
class TensorRes:
    """All the things we pass from the forward pass to the backward pass."""
    mt: MultiTensor
    quant_grad: Union[Callable[[jnp.ndarray], tuple[jnp.ndarray]], None]


@flax.struct.dataclass
class DotGeneralRes:
    context_bwd: Context
    lhs: TensorRes
    rhs: TensorRes


def _scale_trans(x, ca, ba):
    """Transposes x to output dimension order."""
    ca = list(ca)
    ba = list(ba)
    for i in ca:
        assert x.shape[i] == 1
    ra = list(i for i in range(len(x.shape)) if i not in ba + ca)
    x = jnp.transpose(x, ba + ra + ca)
    # TODO(lew): x = jnp.squeeze(x, axis=range(len(ba+ra): len(x.shape))
    shape_ba = x.shape[: len(ba)]
    shape_ra = x.shape[len(ba): len(x.shape) - len(ca)]
    # Will need to add additional axes (size 1) for the other shape_ra
    x = x.reshape(shape_ba + shape_ra)
    return x


def _lhs_scale_transpose(lhs_scale, dimension_numbers, lhs_shape, rhs_shape):
    """Transposes lhs_scale to output dimension order."""
    if lhs_scale is None:
        return None
    # The axis order in out is as follows: batch, lhs_ra, rhs_ra
    # - batch axes order is uniquely determined by either lhs_ba or rhs_ba
    # - contraction axes ca disappear from the output
    # - order of the remaining axes (ra) is preserved.
    (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dimension_numbers
    qlhs_scale_t = _scale_trans(lhs_scale, lhs_ca, lhs_ba)
    # inserting dummy axes for rhs_ra
    assert len(qlhs_scale_t.shape) == len(lhs_shape) - len(lhs_ca)
    start = len(qlhs_scale_t.shape)
    end = len(rhs_shape) - len(rhs_ca) - len(rhs_ba) + start
    lhs_dummy_axes = range(start, end)
    qlhs_scale_t = jnp.expand_dims(qlhs_scale_t, axis=lhs_dummy_axes)
    return qlhs_scale_t


def _rhs_scale_transpose(rhs_scale, dimension_numbers, lhs_shape, rhs_shape):
    if rhs_scale is None:
        return None
    del rhs_shape
    (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dimension_numbers
    qrhs_scale_t = _scale_trans(rhs_scale, rhs_ca, rhs_ba)
    start = len(rhs_ba)
    end = len(lhs_shape) - len(lhs_ca) - len(lhs_ba) + start
    rhs_dummy_axes = range(start, end)
    qrhs_scale_t = jnp.expand_dims(qrhs_scale_t, axis=rhs_dummy_axes)
    return qrhs_scale_t


def _maybe_mul(x, scale):
    if scale is None:
        return x
    return x * scale


def _maybe_inv(x):
    if x is None:
        return None
    return 1.0 / x


def _make_dot_general_raw(cfg: config.DotGeneralRaw):
    """Makes quantized lax.dot_general replacement."""

    msg = 'Custom calib_shared_axes not implemented for local AQT.'
    assert cfg.lhs.calib_shared_axes is None, msg
    assert cfg.rhs.calib_shared_axes is None, msg

    def dot_general_raw(
            lhs: jnp.ndarray,
            rhs: Union[jnp.ndarray, MultiTensor],
            # xhs_qt are used in serving.
            lhs_qt: Optional[QTensor],
            rhs_qt: Optional[QTensor],
            dimension_numbers,
            context,
    ):
        """Creates a dot_general function without custom gradient."""
        (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dimension_numbers
        # We need to copy because we modify cfg to populate some defaults.

        # TODO(lew):
        #  - Use qx.value with the int type.
        #  - Handle qx.value with the int type in an optimized way.
        #  - Add a "FQ" case we multiply qx.value*qx.value_scale (not transposed).
        #  - Can we carry untransposed scale and transpose here?
        if isinstance(rhs, MultiTensor):
            # We are in gradient code.
            fwd_quantized = rhs.qx is not None
            expect_fwd_quantized = cfg.rhs.use_fwd_quant is not None
            msg = (
                'Misconfiguration: use_fwd_quant=True, but there is no fwd'
                ' quantization (but rhs.qx is None).'
            )
            assert fwd_quantized == expect_fwd_quantized, msg
            if cfg.rhs.use_fwd_quant:
                assert rhs.qx is not None, msg
                lhs = _maybe_mul(lhs, rhs.qx.qvalue_scale_t)
                rhs = rhs.qx.qvalue
            else:
                rhs = rhs.x
        else:
            assert cfg.rhs.use_fwd_quant is None, 'cannot set use_fwd_quant in fwd'

        if cfg.local_aqt is not None:

            def factor_reshape(x, ca, ba):
                factor = cfg.local_aqt.contraction_axis_shard_count
                assert factor is not None
                if len(ca) == 0:
                    return x, ca, ba
                shape = list(x.shape)
                ax = ca[0]
                orig_size = shape[ax]
                assert orig_size % factor == 0
                shape[ax] = factor
                shape.insert(ax + 1, orig_size // factor)
                new_ca = [(b + int(b >= ax)) for b in ca]
                assert new_ca[0] == ax + 1
                new_ba = [ax] + [(b + int(b > ax)) for b in ba]
                return x.reshape(shape), new_ca, new_ba

            lhs, lhs_ca, lhs_ba = factor_reshape(lhs, lhs_ca, lhs_ba)
            rhs, rhs_ca, rhs_ba = factor_reshape(rhs, rhs_ca, rhs_ba)

            dimension_numbers = (lhs_ca, rhs_ca), (lhs_ba, rhs_ba)

        assert isinstance(rhs, jnp.ndarray)

        context, context_bwd = _context_split(context)
        context_lhs, context_rhs = _context_split(context)
        del context

        # TODO(lew): Have a function to handle lhs and rhs uniformly.
        if lhs_qt is not None:
            lhs_q, lhs_inv_scale_t = (lhs_qt.qvalue, lhs_qt.qvalue_scale_t)
            lhs_quant_grad = 'Poison. Not needed in serving'
            lhs_inv_scale = 'Poison. Fake quant not used in serving.'
            lhs_qx = lhs_qt
        else:
            lhs_q, lhs_inv_scale, lhs_quant_grad = _scale_quant(
                lhs, cfg=cfg.lhs, ca=lhs_ca, context=context_lhs
            )
            lhs_inv_scale_t = _lhs_scale_transpose(
                lhs_inv_scale, dimension_numbers, lhs.shape, rhs.shape
            )
            lhs_qx = (
                None
                if lhs_inv_scale_t is None
                else QTensor(qvalue=lhs_q, qvalue_scale_t=lhs_inv_scale_t)
            )

        lhs_mt = MultiTensor(x=lhs, qx=lhs_qx)
        lhs_res = TensorRes(mt=lhs_mt, quant_grad=lhs_quant_grad)

        if rhs_qt is not None:
            rhs_q, rhs_inv_scale_t = (rhs_qt.qvalue, rhs_qt.qvalue_scale_t)
            rhs_quant_grad = 'Poison. Not needed in serving'
            rhs_inv_scale = 'Poison. Fake quant not used in serving.'
            rhs_qx = rhs_qt
        else:
            rhs_q, rhs_inv_scale, rhs_quant_grad = _scale_quant(
                rhs, cfg=cfg.rhs, ca=rhs_ca, context=context_rhs
            )
            rhs_inv_scale_t = _rhs_scale_transpose(
                rhs_inv_scale, dimension_numbers, lhs.shape, rhs.shape
            )
            rhs_qx = (
                None
                if rhs_inv_scale_t is None
                else QTensor(qvalue=rhs_q, qvalue_scale_t=rhs_inv_scale_t)
            )
        rhs_mt = MultiTensor(x=rhs, qx=rhs_qx)
        rhs_res = TensorRes(mt=rhs_mt, quant_grad=rhs_quant_grad)

        # TODO(lew): mt.x above should be clipped for clipping calibrations

        # TODO(yichizh): the same code is applied to lhs and rhs.
        # Should make a function of it that includes preprocess as well.
        lhs_cast_dtype = cfg.lhs.numerics.get_dtype()
        rhs_cast_dtype = cfg.rhs.numerics.get_dtype()
        if cfg.lhs.use_fake_quant:
            # TODO(yichizh): replace rounding in numerics with casting to dtype.
            # So fake quant becomes casting to dtype first, then casting to bfloat.
            # This is because FP8 numerics relies on this cast to do the rounding.
            msg = "Can't cast dtype in fake_quant mode."
            assert lhs_cast_dtype is None and rhs_cast_dtype is None, msg
            lhs_q = _maybe_mul(lhs_q, lhs_inv_scale)
            rhs_q = _maybe_mul(rhs_q, rhs_inv_scale)
        else:
            if lhs_cast_dtype is not None:
                lhs_q = lhs_q.astype(lhs_cast_dtype)
            if rhs_cast_dtype is not None:
                rhs_q = rhs_q.astype(rhs_cast_dtype)

        dtype_ms = (
            f'Found {cfg.dg_accumulator_dtype=}, {lhs_cast_dtype=} and'
            f' {rhs_cast_dtype=}. Dot general accumulator dtype can only be'
            ' jnp.int32 when both inputs are int8. Otherwise it is recommended to'
            ' be None to let lax.dot_general automatically decide it.'
        )
        if cfg.dg_accumulator_dtype == jnp.int32:
            assert lhs_cast_dtype == jnp.int8 and rhs_cast_dtype == jnp.int8, dtype_ms
        out = lax.dot_general(
            lhs_q,
            rhs_q,
            dimension_numbers=dimension_numbers,
            preferred_element_type=cfg.dg_accumulator_dtype,
            precision=lax.Precision.DEFAULT,
        ).astype(jnp.promote_types(lhs, rhs))
        # TODO(lew): Do we have a correct precision above?
        #   Relevant: https://github.com/google/jax/issues/14022

        if not cfg.lhs.use_fake_quant:
            out = _maybe_mul(out, lhs_inv_scale_t)
            out = _maybe_mul(out, rhs_inv_scale_t)

        res = DotGeneralRes(
            context_bwd=context_bwd,
            lhs=lhs_res,
            rhs=rhs_res,
        )
        if cfg.local_aqt is not None:
            assert len(lhs_ca) == len(rhs_ca)
            if len(lhs_ca) > 0:
                out = jnp.sum(out, axis=0)
            # We are not supporting local AQT in fwd pass, so no res needed.
            res = None
        return out, res

    return dot_general_raw


def _dot_general_raw_attach_gradient(
        fwd_dot_general_raw,
        dlhs_dot_general_raw,
        drhs_dot_general_raw,
):
    """
    The _dot_general_raw_attach_gradient function is a helper function that attaches the gradient to the dot_general_raw
       function. This allows for backpropagation through the dot general operation. The vjp (vector jacobian product)
       calculates the gradients of each input tensor with respect to each output tensor, and returns them as a tuple in
       order of inputs. In this case, it will return dlhs and drhs which are both TensorRes objects.

    :param fwd_dot_general_raw: Compute the forward pass of the dot general operation
    :param dlhs_dot_general_raw: Compute the gradient of dot_general with respect to lhs
    :param drhs_dot_general_raw: Calculate the gradient of rhs
    :param : Specify the function to be used for forward pass
    :return: A function
    """

    def make_fwd(return_residual):

        def fwd(
                lhs,
                rhs,
                lhs_qt,
                rhs_qt,
                dimension_numbers,
                context,
        ):
            assert lhs.dtype == rhs.dtype
            ret, res = fwd_dot_general_raw(
                lhs,
                rhs,
                lhs_qt,
                rhs_qt,
                dimension_numbers,
                context,
            )
            ret = ret.astype(lhs.dtype)
            # We return these values to allow for materialization.
            qret = (res.lhs.mt.qx, res.rhs.mt.qx)
            if return_residual:
                return ((ret, qret), res)
            else:
                return (ret, qret)

        return fwd

    def vjp_bwd(
            fwd_dimension_numbers,
            res: DotGeneralRes,
            g,
    ):
        # g[1] is gradient with respect to qret which we are ignoring.
        g = g[0]

        def ranges_like(*xs):
            start = 0
            for x in xs:
                yield tuple(range(start, start + len(x)))
                start += len(x)

        def grad_dot_general(
                y_res: TensorRes,
                quant_grad,
                dot_general,
                y_is_lhs,
                context,
        ):
            y_ndim = y_res.mt.x.ndim

            (x_ca, y_ca), (x_ba, y_ba) = fwd_dimension_numbers
            if y_is_lhs:
                (y_ca, x_ca) = (x_ca, y_ca)
                (y_ba, x_ba) = (x_ba, y_ba)
            g_ndim = g.ndim - y_ndim + len(x_ba) + 2 * len(x_ca)
            x_ra = tuple(i for i in range(g_ndim) if i not in x_ca and i not in x_ba)
            y_ra = tuple(i for i in range(y_ndim) if i not in y_ca and i not in y_ba)
            if y_is_lhs:
                g_ba, g_ca, _ = ranges_like(x_ba, y_ra, x_ra)
            else:
                g_ba, _, g_ca = ranges_like(x_ba, x_ra, y_ra)
            dims = ((g_ca, y_ra), (g_ba, y_ba))

            out, _ = dot_general(g, y_res.mt, None, None, dims, context)

            x_ca_sorted_by_y = tuple(onp.take(x_ca, onp.argsort(y_ca)))
            out_axes = tuple(onp.argsort(tuple(x_ba) + x_ra + x_ca_sorted_by_y))
            transposed_out = jax.lax.transpose(out, out_axes)
            if quant_grad is not None:
                transposed_out = quant_grad(transposed_out)[0]
            return transposed_out

        context1, context2 = _context_split(res.context_bwd)
        dlhs = grad_dot_general(
            res.rhs,
            res.lhs.quant_grad,
            dlhs_dot_general_raw,
            False,
            context1,
        )
        drhs = grad_dot_general(
            res.lhs,
            res.rhs.quant_grad,
            drhs_dot_general_raw,
            True,
            context2,
        )
        # fwd_dimension_numbers are marked as nondiff_argnums instead of returning
        # None as grad to it. This is because it is a tuple of Python integers
        # that cannot be traced by Jax.
        return (dlhs, drhs, None, None, None)

    vjp = jax.custom_vjp(make_fwd(False), nondiff_argnums=(4,))
    vjp.defvjp(make_fwd(True), vjp_bwd)

    return vjp


def make_dot_general(cfg: Optional[config.DotGeneral]):

    """
    The make_dot_general function is a wrapper around the dot_general function.
    It takes in two QTensors, lhs and rhs, and returns a QTensor out.
    The make_dot_general function also handles preprocessing of the inputs to dot_general (lhs and rhs)
    and postprocessing of the output from dot_general (out).  The pre-/post-processing steps are:

    :param cfg: Optional[config.DotGeneral]: Specify the configuration of the dot_general operation
    :return: A function that returns a function
    """
    if cfg is None:
        def ret_lax_dg(
                lhs,
                rhs,
                dimension_numbers,
                precision=None,
                preferred_element_type=None,
                *,
                context=Context(key=None, train_step=None),
        ):

            """
            The ret_lax_dg function is a wrapper for the jax.lax.dot_general function,
            which performs a general matrix multiplication of two arrays with batch dimensions
            and/or transpositions applied to either or both inputs. The ret_lax_dg function
            is used in the implementation of the dot operation in this module.

            :param lhs: Specify the left-hand side of the dot product
            :param rhs: Specify the right-hand side of the matrix multiplication
            :param dimension_numbers: Specify the dimensions of the operands
            :param precision: Specify the precision of the computation
            :param preferred_element_type: Specify the type of element that should be used to store the result
            :param *: Indicate that all the following parameters are keyword only
            :param context: Pass in the context of the computation
            :param : Specify the dimension numbers of the dot product
            :return: The result of the dot_general operation
            """
            del context
            return jax.lax.dot_general(
                lhs, rhs, dimension_numbers, precision, preferred_element_type
            )

        return ret_lax_dg

    dg = _dot_general_raw_attach_gradient(
        fwd_dot_general_raw=_make_dot_general_raw(cfg.fwd),
        dlhs_dot_general_raw=_make_dot_general_raw(cfg.dlhs),
        drhs_dot_general_raw=_make_dot_general_raw(cfg.drhs),
    )

    def ret_dg(
            lhs,
            rhs,
            dimension_numbers,
            precision=None,
            preferred_element_type=None,
            *,
            context=Context(key=None, train_step=None),
    ):

        """
        The ret_dg function is a wrapper around the dg function.
        It takes in two QTensors, lhs and rhs, and returns a QTensor out.
        The ret_dg function also handles preprocessing of the inputs to dg (lhs and rhs)
        and postprocessing of the output from dg (out).  The pre-/post-processing steps are:

        :param lhs: Pass the left hand side of the matrix multiplication
        :param rhs: Pass on the right hand side of the matrix multiplication
        :param dimension_numbers: Specify the contraction pattern
        :param precision: Specify the precision of the output
        :param preferred_element_type: Specify the dtype of the output
        :param *: Indicate that the argument is a keyword-only
        :param context: Pass the context to the dg function
        :return: A function that returns a function
        """
        del preferred_element_type
        assert (
                precision is None
        ), f'Precision {precision} requested together with quantization.'

        msg = 'AQT is not yet optimized to accept quantized types directly. '
        msg += f'lhs.dtype: {lhs.dtype}, rhs.dtype: {rhs.dtype}'
        assert lhs.dtype in [jnp.bfloat16, jnp.float32, jnp.float16], msg
        assert rhs.dtype in [jnp.bfloat16, jnp.float32, jnp.float16], msg
        # TODO(lew): Refactor Have a flax class with get and set.
        # TODO(lew): Have a function to handle lhs and rhs uniformly.
        lhs_qt = None
        if cfg.fwd.lhs.preprocess is not None:
            # lhs_q is quantized dtype.
            # we are breaking the invariant that QTensor has a float qvalue
            # But it will just be cast again to the same type.
            lhs_qt = cfg.fwd.lhs.preprocess(None)
        rhs_qt = None
        if cfg.fwd.rhs.preprocess is not None:
            rhs_qt = cfg.fwd.rhs.preprocess(None)

        out, (out_lhs_qt, out_rhs_qt) = dg(
            lhs=lhs,
            rhs=rhs,
            lhs_qt=lhs_qt,
            rhs_qt=rhs_qt,
            dimension_numbers=dimension_numbers,
            context=context,
        )

        if cfg.fwd.lhs.preprocess is not None:
            lhs_dtype = cfg.fwd.lhs.numerics.get_dtype()
            out_lhs_qt = QTensor(
                out_lhs_qt.qvalue.astype(lhs_dtype), out_lhs_qt.qvalue_scale_t
            )
            none = cfg.fwd.lhs.preprocess(out_lhs_qt)
            assert none is None
        if cfg.fwd.rhs.preprocess is not None:
            rhs_dtype = cfg.fwd.rhs.numerics.get_dtype()
            out_rhs_qt = QTensor(
                out_rhs_qt.qvalue.astype(rhs_dtype), out_rhs_qt.qvalue_scale_t
            )
            none = cfg.fwd.rhs.preprocess(out_rhs_qt)
            assert none is None

        return out

    return ret_dg



# Copyright 2023 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module containing rms forward and backward pass."""

import functools

from typing import Optional

import jax
from jax import lax
import jax.numpy as jnp
from jax._src.lax.control_flow.for_loop import for_loop

from jax.experimental import pallas as pl


def rms_norm_forward_kernel(
        x_ref, weight_ref, bias_ref,  # Input arrays
        o_ref, rstd_ref=None,  # Output arrays
        *, eps: float, block_size: int):
    n_col = x_ref.shape[0]

    def var_body(i, acc_ref):
        col_idx = i * block_size + jnp.arange(block_size)
        mask = col_idx < n_col
        a = pl.load(x_ref, (col_idx,), mask=mask, other=0.,
                    eviction_policy="evict_last").astype(jnp.float32)
        a = jnp.where(mask, a, 0.)
        acc_ref[:] += a * a

    var = for_loop(pl.cdiv(n_col, block_size), var_body,
                   jnp.zeros(block_size)).sum() / n_col
    rstd = 1 / jnp.sqrt(var + eps)
    if rstd_ref is not None:
        rstd_ref[...] = rstd.astype(rstd_ref.dtype)

    def body(i, _):
        col_idx = i * block_size + jnp.arange(block_size)
        mask = col_idx < n_col
        weight = pl.load(weight_ref, (col_idx,), mask=mask)
        bias = pl.load(bias_ref, (col_idx,), mask=mask)
        x = pl.load(x_ref, (col_idx,), mask=mask, other=0.,
                    eviction_policy="evict_first").astype(jnp.float32)
        out = x * rstd * weight + bias
        pl.store(o_ref, (col_idx,), out.astype(o_ref.dtype), mask=mask)

    for_loop(pl.cdiv(n_col, block_size), body, ())


def rms_norm_forward(
        x, weight, bias,
        num_warps: Optional[int] = None,
        num_stages: Optional[int] = 3,
        eps: float = 1e-5,
        backward_pass_impl: str = 'triton',
        interpret: bool = False):
    del num_stages
    del backward_pass_impl
    n = x.shape[-1]
    # Triton heuristics
    # Less than 64KB per feature: enqueue fused kernel
    max_fused_size = 65536 // x.dtype.itemsize
    block_size = min(max_fused_size, pl.next_power_of_2(n))
    block_size = min(max(block_size, 128), 4096)
    num_warps = min(max(block_size // 256, 1), 8)

    kernel = functools.partial(rms_norm_forward_kernel, eps=eps,
                               block_size=block_size)
    out_shape = [
        jax.ShapeDtypeStruct(shape=(n,), dtype=x.dtype),
        jax.ShapeDtypeStruct(shape=(), dtype=x.dtype)
    ]
    method = pl.pallas_call(kernel, num_warps=num_warps,
                            grid=(), out_shape=out_shape, debug=False,
                            interpret=interpret, name='rms_forward')

    method = jax.vmap(jax.vmap(method, in_axes=(0, None, None)), in_axes=(0, None, None))
    out, rstd = method(x, weight, bias)
    return out, (x, weight, bias, rstd)


def rms_norm_backward_kernel_dx(
        # Inputs
        x_ref, weight_ref, bias_ref, do_ref,
        rstd_ref,
        # Outputs
        dx_ref,
        *, eps: float, block_size: int):
    n_col = x_ref.shape[0]

    def mean_body(i, c1_acc_ref):
        col_idx = i * block_size + jnp.arange(block_size)
        mask = col_idx < n_col
        a = pl.load(x_ref, (col_idx,), mask=mask, other=0.,
                    eviction_policy="evict_last").astype(jnp.float32)
        dout = pl.load(do_ref, (col_idx,), mask=mask, other=0.,
                       eviction_policy="evict_last").astype(jnp.float32)
        weight = pl.load(weight_ref, (col_idx,), mask=mask, other=0.,
                         eviction_policy="evict_last").astype(jnp.float32)
        a_hat = a * rstd_ref[...]
        wdout = weight * dout
        c1_acc_ref[:] += a_hat * wdout

    c1 = for_loop(pl.cdiv(n_col, block_size), mean_body, jnp.zeros(block_size))
    c1 = c1.sum() / n_col

    def dx_body(i, acc_ref):
        col_idx = i * block_size + jnp.arange(block_size)
        mask = col_idx < n_col
        a = pl.load(x_ref, (col_idx,), mask=mask, other=0.,
                    eviction_policy="evict_last").astype(jnp.float32)
        dout = pl.load(do_ref, (col_idx,), mask=mask, other=0.,
                       eviction_policy="evict_last").astype(jnp.float32)
        weight = pl.load(weight_ref, (col_idx,), mask=mask, other=0.,
                         eviction_policy="evict_last").astype(jnp.float32)
        a_hat = a * rstd_ref[...]
        wdout = weight * dout
        da = (wdout - (a_hat * c1)) * rstd_ref[...]
        pl.store(dx_ref, (col_idx,), da.astype(dx_ref.dtype), mask=mask)

    for_loop(pl.cdiv(n_col, block_size), dx_body, ())


def rms_norm_backward_kernel_dw_db(
        # Inputs
        x_ref, weight_ref, bias_ref, do_ref,
        rstd_ref,
        # Outputs
        dw_ref, db_ref,
        *, eps: float, block_m: int, block_n: int):
    m, n_col = x_ref.shape
    j = pl.program_id(0)
    col_idx = j * block_n + jnp.arange(block_n)
    col_mask = col_idx < n_col

    def body(i, acc_ref):
        row_idx = i * block_m + jnp.arange(block_m)
        row_mask = row_idx < m
        mask = row_mask[:, None] & col_mask[None, :]
        a = pl.load(
            x_ref, (row_idx[:, None], col_idx[None]), mask=mask, other=0.0
        ).astype(jnp.float32)
        dout = pl.load(
            do_ref, (row_idx[:, None], col_idx[None]), mask=mask, other=0.0
        ).astype(jnp.float32)
        rstd = pl.load(rstd_ref, (row_idx,), mask=row_mask, other=0.).astype(jnp.float32)
        a_hat = a * rstd[:, None]
        dw_acc_ref, db_acc_ref = acc_ref
        dw_acc_ref[:] += (dout * a_hat).sum(axis=0)
        db_acc_ref[:] += dout.sum(axis=0)

    dw_acc, db_acc = for_loop(pl.cdiv(m, block_m), body, (jnp.zeros(block_n), jnp.zeros(block_n)))
    pl.store(dw_ref, (col_idx,), dw_acc.astype(dw_ref.dtype), mask=col_mask)
    pl.store(db_ref, (col_idx,), db_acc.astype(db_ref.dtype), mask=col_mask)


def rms_norm_backward(
        num_warps: Optional[int],
        num_stages: Optional[int],
        eps: float,
        backward_pass_impl: str,
        interpret: bool,
        res, do):
    del num_stages
    x, weight, bias, rstd = res
    if backward_pass_impl == 'xla':
        return jax.vjp(rms_norm_reference, x, weight, bias)[1](do)

    *shape_prefix, n = x.shape
    reshaped_x = x.reshape((-1, n))
    reshaped_rstd = rstd.reshape((-1,))
    reshaped_do = do.reshape((-1, n))
    # Triton heuristics
    # Less than 64KB per feature: enqueue fused kernel
    max_fused_size = 65536 // x.dtype.itemsize
    block_size = min(max_fused_size, pl.next_power_of_2(n))
    block_size = min(max(block_size, 128), 4096)
    num_warps = min(max(block_size // 256, 1), 8)

    # rms_norm_backward_kernel_dx parallel over batch dims
    kernel = functools.partial(rms_norm_backward_kernel_dx, eps=eps,
                               block_size=block_size)
    out_shape_dx = jax.ShapeDtypeStruct(shape=(n,), dtype=x.dtype)
    method = pl.pallas_call(kernel, num_warps=num_warps,
                            grid=(), out_shape=out_shape_dx, debug=False,
                            interpret=interpret, name='ln_backward_dx')

    method = jax.vmap(method, in_axes=(0, None, None, 0, 0))
    dx = method(reshaped_x, weight, bias, reshaped_do, reshaped_rstd)
    dx = dx.reshape((*shape_prefix, n))

    # rms_norm_backward_kernel_dw_db reduce over batch dims
    # Triton heuristics
    if n > 10240:
        block_n = 128
        block_m = 32
        num_warps = 4
    else:
        # maximize occupancy for small N
        block_n = 16
        block_m = 16
        num_warps = 8
    kernel = functools.partial(rms_norm_backward_kernel_dw_db, eps=eps,
                               block_m=block_m, block_n=block_n)
    out_shape_dwbias = [
        jax.ShapeDtypeStruct(shape=weight.shape, dtype=weight.dtype),
        jax.ShapeDtypeStruct(shape=bias.shape, dtype=bias.dtype)
    ]
    grid_ = (pl.cdiv(reshaped_x.shape[1], block_n),)
    method = pl.pallas_call(kernel, num_warps=num_warps,
                            grid=grid_, out_shape=out_shape_dwbias, debug=False,
                            interpret=interpret, name='ln_backward_dw_db')
    dw, dbias = method(reshaped_x, weight, bias, reshaped_do, reshaped_rstd)
    return dx, dw, dbias


@functools.partial(jax.custom_vjp, nondiff_argnums=[3, 4, 5, 6, 7])
@functools.partial(jax.jit, static_argnames=["num_warps", "num_stages",
                                             "num_stages", "eps",
                                             "backward_pass_impl",
                                             "interpret"])
def rms_norm(
        x, weight, bias,
        num_warps: Optional[int] = None,
        num_stages: Optional[int] = 3,
        eps: float = 1e-5,
        backward_pass_impl: str = 'triton',
        interpret: bool = False):
    n = x.shape[-1]
    # Triton heuristics
    # Less than 64KB per feature: enqueue fused kernel
    max_fused_size = 65536 // x.dtype.itemsize
    block_size = min(max_fused_size, pl.next_power_of_2(n))
    block_size = min(max(block_size, 128), 4096)
    num_warps = min(max(block_size // 256, 1), 8)

    kernel = functools.partial(rms_norm_forward_kernel, eps=eps,
                               block_size=block_size)
    out_shape = jax.ShapeDtypeStruct(shape=(n,), dtype=x.dtype)
    method = pl.pallas_call(kernel, num_warps=num_warps, num_stages=num_stages,
                            grid=(), out_shape=out_shape, debug=False,
                            interpret=interpret)
    method = jax.vmap(jax.vmap(method, in_axes=(0, None, None)), in_axes=(0, None, None))
    return method(x, weight, bias)


rms_norm.defvjp(rms_norm_forward, rms_norm_backward)


@functools.partial(jax.jit, static_argnames=["eps"])
@functools.partial(jax.vmap, in_axes=(0, None, None), out_axes=0)
def rms_norm_reference(x, weight, bias, *, eps: float = 1e-5):
    var = jnp.mean(jnp.square(x), axis=1)
    mul = lax.rsqrt(var + eps)
    return x * mul[:, None] * weight[None] + bias[None]



import chex
import jax
from jax import numpy as jnp


def global_norm(tree):
    """ Return the global L2 norm of a pytree. """
    squared = jax.tree_util.tree_map(lambda x: jnp.sum(jnp.square(x)), tree)
    flattened, _ = jax.flatten_util.ravel_pytree(squared)
    return jnp.sqrt(jnp.sum(flattened))


def average_metrics(metrics):
    """
    The average_metrics function takes a list of metrics and averages them.

    :param metrics: Store the metrics for each batch
    :return: The mean of the metrics across all runs
    
    """
    return jax.tree_map(
        lambda *args: jnp.mean(jnp.stack(args)),
        *metrics
    )


def transpose(array: chex.Array, dim0: int, dim1: int):
    """
    The transpose function takes an array and two dimensions, and returns a new
    array with the specified dimensions transposed. The first dimension is given as
    a positive integer, where 0 represents the outermost dimension of the array. If
    the first dimension is negative, it counts from the end of the shape tuple; -2
    is equivalent to len(shape) - 2. The second dimension may be specified in a similar way.

    :param array: chex.Array: Specify the array to be transposed
    :param dim0: int: Specify the first dimension to be transposed
    :param dim1: int: Specify the dimension of the array
    :return: A new array with the same data, but with axes permuted
    
    """
    dim0 = dim0 if dim0 > 0 else array.ndim - dim0
    dim1 = dim1 if dim1 > 0 else array.ndim - dim1
    perm = list(range(array.ndim))
    perm[dim0], perm[dim1] = perm[dim1], perm[dim0]
    return jnp.transpose(array, perm)


def fused_softmax(x: chex.Array, axis: int = -1):
    """
    The fused_softmax function is a fused version of the softmax function.

    :param x: chex.Array: Specify the input to the function
    :param axis: int: Specify the axis along which to apply the softmax function
    :return: The same result as the softmax function
    
    """
    return jnp.exp(jax.nn.log_softmax(x, axis=axis))



import os
import jax
import flax
import tqdm
from flax.serialization import (
    from_bytes, to_bytes, to_state_dict, from_state_dict
)
from flax.traverse_util import flatten_dict, unflatten_dict, empty_node
import msgpack
from jax import numpy as jnp

from flax import struct
from typing import Callable, Literal


def get_dtype(tensor, dtype):
    if dtype is None or dtype == "":
        return tensor
    if isinstance(dtype, str):
        dtype = {
            "bf16": jnp.bfloat16,
            "bfloat16": jnp.bfloat16,
            "fp16": jnp.float16,
            "float16": jnp.float16,
            "fp32": jnp.float32,
            "float32": jnp.float32,
            "fp64": jnp.float64,
            "float64": jnp.float64,
        }[dtype]
    float_dtypes = (jnp.bfloat16, jnp.float16, jnp.float32, jnp.float64)
    if getattr(tensor, "dtype", None) in float_dtypes:
        tensor = tensor.astype(dtype)
    return tensor


class CheckpointManager(object):
    """
    Custom msgpack checkpointer that saves large train states by serializing
    and saving tensors one by one in a streaming fashion. Avoids running
    out of memory or local TPU disk with default flax checkpointer.
    """

    def __init__(
            self,
            checkpoint_dir,
            enable=True,
            float_dtype: str | jnp.dtype = "bf16",
            save_optimizer_state: bool = True,
            verbose: bool = False
    ):
        self.float_dtype = float_dtype
        self.save_optimizer_state = save_optimizer_state
        self.checkpoint_dir = checkpoint_dir
        self.enable = enable
        self.verbose = verbose

    def save_checkpoint(
            self,
            state: struct.PyTreeNode,
            filename: str | os.PathLike,
            gather_fns: dict[Callable] = None,
            mismatch_allowed: bool = True

    ):
        if self.enable:
            path = os.path.join(self.checkpoint_dir, filename)
        else:
            path = "/dev/null"
        self.save_state_to_file(
            state, path, gather_fns, self.float_dtype, mismatch_allowed=mismatch_allowed
        )

    @staticmethod
    def save_state_to_file(
            state: struct.PyTreeNode,
            path: str | os.PathLike,
            gather_fns: dict[Callable] = None,
            float_dtype=None,
            verbose: bool = False,
            mismatch_allowed: bool = True
    ):
        state = to_state_dict(state)
        packer = msgpack.Packer()
        flatten_state = flatten_dict(state)
        if gather_fns is not None:
            gather_fns = flatten_dict(to_state_dict(gather_fns))
        pbar = tqdm.tqdm(
            flatten_state.items(),
            disable=not verbose,
            desc="Saving State to File",
        )

        gather_functions_mismatch = 0

        with open(path, "wb") as stream:
            for key, value in pbar:
                if gather_fns is not None:
                    try:
                        callable_func = gather_fns[key]
                        if callable_func is None and not mismatch_allowed:
                            raise KeyError(f"Gather Function {key} is None and NoneType OBJ is not callable.")
                        value = callable_func(value) if callable_func is not None else value
                        if callable_func is None:
                            gather_functions_mismatch += 1
                    except KeyError as k_err:
                        if mismatch_allowed:
                            gather_functions_mismatch += 1
                        else:
                            raise KeyError(k_err)
                pbar.set_postfix(gather_functions_mismatch=gather_functions_mismatch)
                value = get_dtype(value, float_dtype)
                stream.write(packer.pack((key, to_bytes(value))))

    def save_pickle(
            self,
            obj,
            filename: str | os.PathLike
    ):
        """
        The save_pickle function saves a Python object to disk using the pickle module.

        :param self: Represent the instance of the class
        :param obj: Pass the object that is to be pickled
        :param filename: Specify the name of the file to be saved
        :return: A pickle object
        
        """
        import pickle

        def save_pickle(obj_, path_):
            with open(path_, "wb") as stream:
                pickle.dump(obj_, stream)

        if self.enable:
            path = os.path.join(self.checkpoint_dir, filename)
        else:
            path = "/dev/null"
        save_pickle(obj, path)

    def save_all(
            self,
            state: struct.PyTreeNode,
            gather_fns,
            metadata=None,
            dataset=None,
            milestone=False
    ):
        """
        The save_all function saves the following:
            - metadata.pkl (a pickle file containing a dictionary of metadata)
            - dataset.pkl (a pickle file containing the training data)
            - streaming_params_{step}.pkl or streaming_state_{step}.pkl
                (depending on whether we want to save optimizer state or not,
                this is a checkpoint that will not be overwritten by future checkpoints)

        :param self: Access the attributes and methods of the class
        :param state: struct.PyTreeNode: Save the current state of the model
        :param gather_fns: Gather the state of the optimizer
        :param metadata: Save the metadata of the training
        :param dataset: Save the dataset to disk
        :param milestone: Determine whether the checkpoint is a milestone or not
        :return: Nothing
        
        """
        step = int(jax.device_get(state.step))
        if self.save_optimizer_state:
            checkpoint_state = state
            checkpoint_name = "streaming_state"
            checkpoint_gather_fns = gather_fns
        else:
            checkpoint_state = state.params["params"]
            checkpoint_name = "streaming_params"
            checkpoint_gather_fns = gather_fns.params["params"]

        if milestone:
            # Save a milestone checkpoint that will not be overwritten
            self.save_pickle(metadata, f"metadata_{step}.pkl")
            self.save_pickle(dataset, f"dataset_{step}.pkl")
            self.save_checkpoint(
                checkpoint_state, f"{checkpoint_name}_{step}", checkpoint_gather_fns
            )
        else:
            # Save a normal checkpoint that can be overwritten
            self.save_pickle(metadata, "metadata.pkl")
            self.save_pickle(dataset, "dataset.pkl")
            self.save_checkpoint(
                checkpoint_state, f"{checkpoint_name}", checkpoint_gather_fns
            )

    @staticmethod
    def load_checkpoint(
            path: str | os.PathLike,
            target=None,
            shard_fns: dict[Callable] = None,
            remove_dict_prefix=None,
            verbose: bool = False,
            mismatch_allowed: bool = True,
    ):
        """
        The load_checkpoint function is used to checkpoint a checkpoint from disk.

        :param path: Specify the path to the checkpoint file
        :param target: Specify the model to checkpoint the checkpoint into
        :param shard_fns: Specify a function that will be applied to each tensor in the checkpoint
        :param remove_dict_prefix: Remove the prefix of a dictionary     
        :param verbose: print state and other stuff
        :param mismatch_allowed: when ever to allow shard_fns to be passed even if their None
        :return:  of the form {key: value}, where key is a tuple and value is a tensor
        
        """
        if shard_fns is not None:
            shard_fns = flatten_dict(
                to_state_dict(shard_fns)
            )
        if remove_dict_prefix is not None:
            remove_dict_prefix = tuple(remove_dict_prefix)
        flatten_state = {}

        shard_functions_mismatch = 0
        with open(path, "rb") as fin:
            unpacker = msgpack.Unpacker(fin, read_size=83886080, max_buffer_size=0)
            pbar = tqdm.tqdm(
                unpacker,
                disable=not verbose,
                desc="Loading Checkpoints From File"
            )
            for key, value in pbar:
                key = tuple(key)
                if remove_dict_prefix is not None:
                    if key[:len(remove_dict_prefix)] == remove_dict_prefix:
                        key = key[len(remove_dict_prefix):]
                    else:
                        continue

                tensor = from_bytes(None, value)
                if shard_fns is not None:
                    try:
                        callable_func = shard_fns[key]
                        if callable_func is None and not mismatch_allowed:
                            raise KeyError(f"Shard Function {key} is None and NoneType OBJ is not callable.")
                        tensor = callable_func(tensor) if callable_func is not None else tensor
                        if callable_func is None:
                            shard_functions_mismatch += 1
                    except KeyError as k_err:
                        if mismatch_allowed:
                            shard_functions_mismatch += 1
                        else:
                            raise KeyError(k_err)
                flatten_state[key] = tensor
                pbar.set_postfix(shard_functions_mismatch=shard_functions_mismatch)
        if target is not None:
            flattened_target = flatten_dict(
                to_state_dict(target), keep_empty_nodes=True
            )
            for key, value in flattened_target.items():
                if key not in flatten_state and value == empty_node:
                    flatten_state[key] = value

        state = unflatten_dict(flatten_state)
        if target is None:
            return state

        return from_state_dict(target, state)

    @staticmethod
    def load_flax_checkpoint(
            path,
            target=None,
            shard_fns=None
    ):
        """ Load a standard flax checkpoint that"s not saved with the
            msgpack streaming format.
        """
        with open(path, "rb") as fin:
            encoded_bytes = fin.read()

        state_dict = flax.serialization.msgpack_restore(encoded_bytes)
        if shard_fns is not None:
            shard_fns = to_state_dict(shard_fns)
            state_dict = jax.tree_util.tree_map(lambda fn, x: fn(x), shard_fns, state_dict)

        if target is None:
            return state_dict
        return from_state_dict(target, state_dict)

    @classmethod
    def load_state_checkpoint(
            cls,
            load_type: Literal[
                "state",
                "state_params",
                "params",
                "flax_params"
            ],
            load_path: str | os.PathLike,
            state_target=None,
            state_shard_fns=None,
            disallow_state=False,
            mismatch_allowed: bool = True
    ):
        """
        The load_state_checkpoint function is used to checkpoint a checkpoint from disk.

        :param cls: Call the load_checkpoint function
        :param load_type: Specify which part of state to checkpoint
        :param load_path: Specify where to checkpoint the model from
        :param state_target: Specify the target for the train state
        :param state_shard_fns: Specify the sharding function
        :param disallow_state: Prevent loading the entire state
        :param mismatch_allowed: when ever to allow shard func to be None
        :return: A tuple of two objects, the state and restored_params
        
        """
        if state_target is not None:
            params_target = state_target.params["params"]
        else:
            params_target = None

        if state_shard_fns is not None:
            params_shard_fns = state_shard_fns.params["params"]
        else:
            params_shard_fns = None

        if disallow_state:
            assert load_type != "state", "Loading full state is not allowed!"
        state = None
        restored_params = None
        if load_type == "state":
            state = cls.load_checkpoint(
                path=load_path,
                target=state_target,
                shard_fns=state_shard_fns,
                mismatch_allowed=mismatch_allowed
            )
        elif load_type == "state_params":
            restored_params = cls.load_checkpoint(
                path=load_path,
                target=params_target,
                shard_fns=params_shard_fns,
                remove_dict_prefix=("params", "params"),
                mismatch_allowed=mismatch_allowed
            )
            restored_params = flax.core.frozen_dict.freeze(
                {"params": restored_params}
            )
        elif load_type == "params":
            restored_params = cls.load_checkpoint(
                path=load_path,
                target=params_target,
                shard_fns=params_shard_fns,
                mismatch_allowed=mismatch_allowed
            )
            restored_params = flax.core.frozen_dict.freeze(
                {"params": restored_params}
            )
        elif load_type == "flax_params":
            restored_params = cls.load_flax_checkpoint(
                path=load_path,
                target=params_target,
                shard_fns=params_shard_fns,
            )
            restored_params = flax.core.frozen_dict.freeze(
                {"params": restored_params}
            )
        else:
            raise ValueError(f"Invalid load_from type: {load_type}")

        return state, restored_params



# Copyright 2023 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module containing fused attention forward and backward pass."""
from __future__ import annotations

import dataclasses
import functools
from typing import Any, NamedTuple

import jax
from jax import lax
from jax.experimental import pallas as pl
from jax.experimental.pallas import tpu as pltpu
import jax.numpy as jnp

DEFAULT_MASK_VALUE = -0.7 * float(jnp.finfo(jnp.dtype("float32")).max)
NUM_LANES = 128
NUM_SUBLANES = 8


class SegmentIds(NamedTuple):
    """SegmentIds for Q and KV sequences.

    SegmentIds are used to generate segment mask, which prevents attention between
    different segments in the input sequence. Each array is a list of ids
    (integers).
    Only the token with the same id can attend to each other.

    Attributes:
      q: segment ids along the Q sequence.
      kv: segment ids along the KV sequence.
    """

    q: jax.Array  # [q_seq_len]
    kv: jax.Array  # [kv_seq_len]


@dataclasses.dataclass(frozen=True)
class BlockSizes:
    """Tile sizes parameterizing FlashAttention kernels.

    Those parameters have negligible effect on numerics, but affect performance
    greatly.
    """
    block_q: int
    block_k_major: int
    block_k: int
    block_b: int

    block_q_major_dkv: int | None = None
    block_k_major_dkv: int | None = None
    block_k_dkv: int | None = None
    block_q_dkv: int | None = None

    block_k_major_dq: int | None = None
    block_k_dq: int | None = None
    block_q_dq: int | None = None

    def __post_init__(self):
        def verify_major_minor(prefix, suffix, major, minor):
            if minor > major:
                raise ValueError(
                    f"{prefix}{suffix}={minor} should be smaller than"
                    f" {prefix}_major{suffix}={major}"
                )
            if major % minor != 0:
                raise ValueError(
                    f"{prefix}{suffix}={minor} should divide"
                    f" {prefix}_major{suffix}={major}"
                )

        verify_major_minor("block_k", "", self.block_k_major, self.block_k)
        if self.block_q_major_dkv is not None and self.block_q_dkv is not None:
            verify_major_minor(
                "block_q", "_dkv", self.block_q_major_dkv, self.block_q_dkv
            )
        if self.block_k_major_dkv is not None and self.block_k_dkv is not None:
            verify_major_minor(
                "block_k", "_dkv", self.block_k_major_dkv, self.block_k_dkv
            )
        if self.block_k_major_dq is not None and self.block_k_dq is not None:
            verify_major_minor(
                "block_k", "_dq", self.block_k_major_dq, self.block_k_dq
            )

    @property
    def has_backward_blocks(self) -> bool:
        backward_blocks = (
            self.block_q_major_dkv,
            self.block_k_major_dkv,
            self.block_q_dkv,
            self.block_k_dkv,
            self.block_k_major_dq,
            self.block_k_dq,
            self.block_q_dq,
        )
        return all(b is not None for b in backward_blocks)

    @classmethod
    def get_default(cls, batch_size, num_heads, q_seq_len, kv_len, d_model):
        # TODO(apaszke,sharadmv): Select better parameters based on a heuristic.
        del batch_size, num_heads, q_seq_len, kv_len, d_model  # Unused.
        return BlockSizes(
            block_q=128,
            block_k_major=128,
            block_k=128,
            block_b=1,
            block_q_major_dkv=128,
            block_k_major_dkv=128,
            block_k_dkv=128,
            block_q_dkv=128,
            block_k_major_dq=128,
            block_k_dq=128,
            block_q_dq=128,
        )


def _flash_attention(
        q,
        k,
        v,
        carry,
        q_chunk_idx_start,
        k_chunk_idx_start,
        ab,
        segment_ids,
        save_residuals,
        causal,
        sm_scale,
        block_sizes,
        debug,
):
    return _flash_attention_impl(
        q,
        k,
        v,
        carry,
        q_chunk_idx_start,
        k_chunk_idx_start,
        ab,
        segment_ids,
        save_residuals,
        causal,
        sm_scale,
        block_sizes.block_b,
        block_sizes.block_q,
        block_sizes.block_k_major,
        block_sizes.block_k,
        debug,
    )


def _flash_attention_fwd(
        q,
        k,
        v,
        carry,
        q_chunk_idx_start,
        k_chunk_idx_start,
        ab,
        segment_ids,
        save_residuals,
        causal,
        sm_scale,
        block_sizes,
        debug,
):
    if save_residuals:
        raise NotImplementedError("Higher-order AD not supported")
    assert segment_ids is None  # TODO correct q/k_chunk_idx_start for these
    o, l, m = _flash_attention(
        q, k, v, carry, q_chunk_idx_start, k_chunk_idx_start,
        ab, segment_ids, True, causal, sm_scale, block_sizes, debug
    )
    return o, l, m


def _flash_attention_bwd(
        save_residuals: bool,
        causal: bool,
        sm_scale: float,
        block_sizes: BlockSizes,
        debug: bool,
        q_chunk_idx_start,
        k_chunk_idx_start,
        residuals,
        do,
):
    """VJP rule for FlashAttention."""
    if save_residuals:
        raise NotImplementedError("Higher-order AD not supported")
    (q, k, v, ab, segment_ids, o, l, m) = residuals
    if not block_sizes.has_backward_blocks:
        raise ValueError(
            "Program is being differentiated, but not all backward blocks are"
            " specified"
        )

    di = jnp.sum(
        o.astype(jnp.float32) * do.astype(jnp.float32), axis=-1
    )  # [batch_size, num_heads, q_seq_len]

    dk, dv = _flash_attention_bwd_dkv(
        q_chunk_idx_start,
        k_chunk_idx_start,
        q,
        k,
        v,
        ab,
        segment_ids,
        l,
        m,
        do,
        di,
        block_q_major=block_sizes.block_q_major_dkv,
        block_k_major=block_sizes.block_k_major_dkv,
        block_k=block_sizes.block_k_dkv,
        block_q=block_sizes.block_q_dkv,
        sm_scale=sm_scale,
        causal=causal,
        mask_value=DEFAULT_MASK_VALUE,
        debug=debug,
    )

    dq, ds = _flash_attention_bwd_dq(
        q_chunk_idx_start,
        k_chunk_idx_start,
        q,
        k,
        v,
        ab,
        segment_ids,
        l,
        m,
        do,
        di,
        block_q_major=block_sizes.block_q_dq,
        block_k_major=block_sizes.block_k_major_dq,
        block_k=block_sizes.block_k_dq,
        sm_scale=sm_scale,
        causal=causal,
        mask_value=DEFAULT_MASK_VALUE,
        debug=debug,
    )
    return dq, dk, dv  # , ds, None


MIN_BLOCK_SIZE = 128
TRANS_B_DIM_NUMBERS = (((1,), (1,)), ((), ()))


def below_or_on_diag(r, r_blk_size, c, c_blk_size):
    # A block is considered below or on diagonal as long as the bottom left
    # corner of the block is below or on diagonal.
    return ((r + 1) * r_blk_size - 1) > (c * c_blk_size)


def _flash_attention_kernel(q_idx_chunk_start, k_idx_chunk_start, q_tile_ref, *args, **kwargs):
    block_b = q_tile_ref.shape[0]
    # If we're not going to tile the softmax, then we can avoid a bunch of VPU ops.
    if kwargs["block_k"] == kwargs["kv_seq_len"]:
        assert False
        kernel = _flash_attention_kernel_single_batch_single_step
    else:
        kernel = _flash_attention_kernel_single_batch
    for batch_idx in range(block_b):
        kernel((batch_idx, 0), q_idx_chunk_start, k_idx_chunk_start, q_tile_ref, *args, **kwargs)


def _flash_attention_kernel_single_batch(
        batch_idx: tuple[int, ...],
        q_chunk_idx_start_ref,
        k_chunk_idx_start_ref,
        q_tile_ref,
        k_tile_ref,
        v_tile_ref,
        acc_tile_ref,
        l_tile_ref,
        m_tile_ref,
        ab_tile_ref,
        q_segment_ids_tile_ref,
        kv_segment_ids_tile_ref,  # Input arrays
        o_tile_ref,  # Output arrays
        m_scratch_ref,
        l_scratch_ref,
        acc_scratch_ref,
        l_ref: Any | None = None,
        m_ref: Any | None = None,
        *,
        causal,
        sm_scale,
        block_k,
        kv_seq_len,
        mask_value,
        block_q,
):
    block_k_major = k_tile_ref.shape[2]
    block_q = q_tile_ref.shape[2]
    head_dim = q_tile_ref.shape[-1]

    kv_seq_idx = pl.program_id(3)

    @pl.when(kv_seq_idx == 0)
    def start_new_sequence():
        m_scratch_ref[batch_idx] = m_tile_ref[batch_idx]
        l_scratch_ref[batch_idx] = l_tile_ref[batch_idx]
        acc_scratch_ref[batch_idx] = acc_tile_ref[batch_idx]

    q_chunk_idx_start = q_chunk_idx_start_ref[0]
    k_chunk_idx_start = k_chunk_idx_start_ref[0]

    q_seq_idx = pl.program_id(2)
    if causal:
        should_run = below_or_on_diag(q_seq_idx + q_chunk_idx_start, block_q, kv_seq_idx + k_chunk_idx_start,
                                      block_k_major)
    else:
        should_run = True

    @pl.when(should_run)
    def run():
        @functools.partial(
            lax.fori_loop, 0, block_k_major // block_k, init_val=None
        )
        def body(i, _):
            m_prev = m_scratch_ref[batch_idx]
            l_prev = l_scratch_ref[batch_idx]
            q = q_tile_ref[batch_idx]  # [block_q, head_dim]
            start_k = i * block_k
            k = pl.load(
                k_tile_ref, (*batch_idx, pl.dslice(start_k, block_k), slice(None))
            )  # [block_k, head_dim]

            s = jax.lax.dot_general(
                q, k, TRANS_B_DIM_NUMBERS, preferred_element_type=jnp.float32
            )  # [block_q, block_k]

            # Add attention bias if needed.
            if ab_tile_ref is not None:
                ab = pl.load(
                    ab_tile_ref,
                    (batch_idx[0], pl.dslice(start_k, block_k))
                ).astype(jnp.float32)
                s += ab

            if sm_scale != 1.0:
                s *= sm_scale

            mask = None
            if q_segment_ids_tile_ref is not None:
                repeats, rem = divmod(block_k, NUM_LANES)
                if rem:
                    raise NotImplementedError(
                        f"kv block size must be a multiple of {NUM_LANES}"
                    )
                q_segment_ids = pltpu.repeat(
                    q_segment_ids_tile_ref[batch_idx[0]], repeats, axis=1
                )  # [block_q, block_k].
                kv_segment_ids = pl.load(
                    kv_segment_ids_tile_ref,
                    (batch_idx[0], pl.dslice(1), pl.dslice(start_k, block_k)),
                )  # [1, block_k].
                mask = jnp.equal(q_segment_ids, kv_segment_ids).astype(jnp.bool_)

            if causal:
                mask_shape = (block_q, block_k)
                row_ids = jax.lax.broadcasted_iota(jnp.int32, mask_shape, 0)
                row_ids += (q_seq_idx + q_chunk_idx_start) * block_q
                col_ids = jax.lax.broadcasted_iota(jnp.int32, mask_shape, 1)
                col_ids += (kv_seq_idx + k_chunk_idx_start) * block_k_major + start_k
                causal_mask = col_ids <= row_ids
                mask = (
                    causal_mask if mask is None else jnp.logical_and(mask, causal_mask)
                )

            s = s if mask is None else s + jnp.where(mask, 0.0, mask_value)

            m_curr = jnp.max(s, axis=1)[:, None]  # Row max, shape [block_q, 1].
            m_next = jnp.maximum(m_prev, m_curr)  # Shape [block_q, 128].

            block_k_repeats, rem = divmod(block_k, MIN_BLOCK_SIZE)
            if rem:
                raise NotImplementedError(
                    f"{block_k=} should be a multiple of {MIN_BLOCK_SIZE}"
                )
            p = jnp.exp(s - pltpu.repeat(m_next, block_k_repeats, 1))

            alpha = jnp.exp(m_prev - m_next)  # Shape [block_q, 128].

            l_corr = alpha * l_prev

            l_next = jnp.sum(p, axis=1)[:, None] + l_corr  # Shape [block_q, 128]

            head_dim_repeats, rem = divmod(head_dim, MIN_BLOCK_SIZE)
            l_broadcast = lambda l: pltpu.repeat(l, head_dim_repeats, 1)
            if rem:
                if head_dim_repeats == 0:
                    l_broadcast = lambda l: l[:, :head_dim]
                else:
                    raise NotImplementedError(
                        f"{head_dim=} should be a multiple of {MIN_BLOCK_SIZE} if larger"
                    )
            l_scratch_ref[batch_idx] = l_next
            m_scratch_ref[batch_idx] = m_next

            l_next_inv_safe = jnp.where(l_next == 0.0, 1.0, 1.0 / l_next)
            acc_scratch_ref[batch_idx] *= l_broadcast(l_corr * l_next_inv_safe)
            v = pl.load(
                v_tile_ref, (*batch_idx, pl.dslice(start_k, block_k), slice(None))
            )
            o_curr = jax.lax.dot(
                p.astype(v.dtype), v, preferred_element_type=jnp.float32
            )
            acc_scratch_ref[batch_idx] += o_curr * l_broadcast(l_next_inv_safe)

    @pl.when(kv_seq_idx == (kv_seq_len // block_k_major) - 1)
    def store_output():
        o_tile_ref[batch_idx] = acc_scratch_ref[batch_idx].astype(o_tile_ref.dtype)
        if l_ref is not None:
            l_ref[batch_idx] = l_scratch_ref[batch_idx].astype(l_ref.dtype)
        if m_ref is not None:
            m_ref[batch_idx] = m_scratch_ref[batch_idx].astype(m_ref.dtype)


def _flash_attention_impl(
        q,
        k,
        v,
        carry,
        q_chunk_idx_start,
        k_chunk_idx_start,
        ab,
        segment_ids,
        save_residuals,
        causal,
        sm_scale,
        block_b,
        block_q,
        block_k_major,
        block_k,
        debug,
):
    assert block_k_major == block_k, (block_k_major, block_k)
    batch_size, num_heads, q_seq_len, head_dim = q.shape
    _, _, kv_seq_len, _ = k.shape
    acc, l_prev, m_prev = carry
    l_prev, m_prev = map(lambda x: jnp.broadcast_to(x[..., None], (*x.shape, MIN_BLOCK_SIZE)), (l_prev, m_prev))
    q_chunk_idx_start, k_chunk_idx_start = q_chunk_idx_start[None], k_chunk_idx_start[None]
    _verify_block("block_q", "q_seq_len", block_q, q_seq_len, should_divide=False)
    _verify_block("block_k_major", "kv_seq_len", block_k_major, kv_seq_len)
    _verify_block("block_k", "kv_seq_len", block_k, kv_seq_len)
    _verify_block("block_b", "batch", block_b, batch_size, should_divide=False)

    # TODO(apaszke): Tile over heads as well.
    grid = (
        pl.cdiv(batch_size, block_b),
        num_heads,
        pl.cdiv(q_seq_len, block_q),
        kv_seq_len // block_k_major,
    )

    def q_index_map(batch_index, head_index, q_seq_index, _, q_idx_ref, k_idx_ref):
        return (batch_index, head_index, q_seq_index, 0)

    def kv_index_map(batch_index, head_index, q_seq_index, kv_seq_index, q_idx_ref, k_idx_ref):
        if causal:
            # If the kv block is skipped, prefetch the next valid kv block, i.e. the
            # 0th one to be used for the next block_q rows.
            next_kv_index = lax.select(
                below_or_on_diag(q_seq_index + q_idx_ref[0], block_q, kv_seq_index + k_idx_ref[0], block_k_major),
                kv_seq_index,
                0,
            )
        else:
            next_kv_index = kv_seq_index
        return (batch_index, head_index, next_kv_index, 0)

    def ab_index_map(batch_index, head_index, q_seq_index, kv_seq_index, q_idx_ref, k_idx_ref):
        if causal:
            should_run = below_or_on_diag(
                q_seq_index + q_idx_ref[0], block_q, kv_seq_index + k_idx_ref[0], block_k_major
            )
            next_kv_index = lax.select(should_run, kv_seq_index, 0)
        else:
            next_kv_index = kv_seq_index

        return (batch_index, next_kv_index)

    def o_index_map(batch_index, head_index, q_seq_index, _, q_idx_ref, k_idx_ref):
        return (batch_index, head_index, q_seq_index, 0)

    def lm_index_map(batch_index, head_index, q_seq_index, _, q_idx_ref, k_idx_ref):
        return (batch_index, head_index, q_seq_index, 0)

    kernel = functools.partial(
        _flash_attention_kernel,
        causal=causal,
        mask_value=DEFAULT_MASK_VALUE,
        sm_scale=sm_scale,
        block_k=block_k,
        kv_seq_len=kv_seq_len,
        block_q=block_q,
    )
    out_shape = [jax.ShapeDtypeStruct(shape=q.shape, dtype=q.dtype)]
    out_specs = [pl.BlockSpec(o_index_map, (block_b, 1, block_q, head_dim))]

    if block_k != kv_seq_len:
        scratch_shape = functools.partial(jax.ShapeDtypeStruct, dtype=jnp.float32)
        m_scratch = scratch_shape((block_b, 1, block_q, MIN_BLOCK_SIZE))
        l_scratch = scratch_shape((block_b, 1, block_q, MIN_BLOCK_SIZE))
        acc_scratch = scratch_shape((block_b, 1, block_q, head_dim))
        out_shape += [m_scratch, l_scratch, acc_scratch]
        out_specs += [
            pl.BlockSpec(lambda *_: (0, 0, 0, 0), m_scratch.shape),
            pl.BlockSpec(lambda *_: (0, 0, 0, 0), l_scratch.shape),
            pl.BlockSpec(lambda *_: (0, 0, 0, 0), acc_scratch.shape),
        ]
    else:
        assert False
        out_shape += [None, None, None]
        out_specs += [None, None, None]

    if save_residuals:
        out_specs = [
            *out_specs,
            pl.BlockSpec(lm_index_map, (block_b, 1, block_q, MIN_BLOCK_SIZE)),
            pl.BlockSpec(lm_index_map, (block_b, 1, block_q, MIN_BLOCK_SIZE)),
        ]
        l = jax.ShapeDtypeStruct(
            (batch_size, num_heads, q_seq_len, MIN_BLOCK_SIZE), dtype=jnp.float32
        )
        m = jax.ShapeDtypeStruct(
            (batch_size, num_heads, q_seq_len, MIN_BLOCK_SIZE), dtype=jnp.float32
        )
        out_shape = (*out_shape, l, m)

    ab_block_spec = (
        pl.BlockSpec(ab_index_map, (block_b, block_k_major))
        if ab is not None else None)

    q_segment_ids_spec = kv_segment_ids_spec = None
    q_segment_ids = kv_segment_ids = None
    if segment_ids is not None:
        assert False

        def q_segment_ids_index_map(batch_index, head_index, q_seq_index, _):
            del head_index
            return (batch_index, q_seq_index, 0)

        def kv_segment_ids_index_map(
                batch_index, head_index, q_seq_index, kv_seq_index
        ):
            del head_index
            if causal:
                next_kv_index = lax.select(
                    below_or_on_diag(q_seq_index, block_q, kv_seq_index, block_k_major),
                    kv_seq_index,
                    0,
                )
            else:
                next_kv_index = kv_seq_index
            return (batch_index, 0, next_kv_index)

        q_segment_ids_spec = pl.BlockSpec(
            q_segment_ids_index_map, (block_b, block_q, NUM_LANES)
        )
        kv_segment_ids_spec = pl.BlockSpec(
            kv_segment_ids_index_map, (block_b, NUM_SUBLANES, block_k_major)
        )

        q_segment_ids = jax.lax.broadcast_in_dim(
            segment_ids.q,
            (batch_size, q_seq_len, NUM_LANES),
            (
                0,
                1,
            ),
        )
        kv_segment_ids = jax.lax.broadcast_in_dim(
            segment_ids.kv,
            (batch_size, NUM_SUBLANES, kv_seq_len),
            (
                0,
                2,
            ),
        )

    in_specs = [
        pl.BlockSpec(q_index_map, (block_b, 1, block_q, head_dim)),
        pl.BlockSpec(kv_index_map, (block_b, 1, block_k_major, head_dim)),
        pl.BlockSpec(kv_index_map, (block_b, 1, block_k_major, head_dim)),
        pl.BlockSpec(q_index_map, (block_b, 1, block_q, head_dim)),
        pl.BlockSpec(lm_index_map, (block_b, 1, block_q, MIN_BLOCK_SIZE)),
        pl.BlockSpec(lm_index_map, (block_b, 1, block_q, MIN_BLOCK_SIZE)),
        ab_block_spec,
        q_segment_ids_spec,
        kv_segment_ids_spec,
    ]

    o, *aux = pl.pallas_call(
        kernel,
        out_shape=out_shape,
        grid_spec=pltpu.PrefetchScalarGridSpec(
            num_scalar_prefetch=2,
            in_specs=in_specs,
            out_specs=out_specs,
            grid=grid
        ),
        debug=debug,
        mosaic_params=dict(
            dimension_semantics=("parallel", "parallel", "parallel", "arbitrary")
        ),
    )(q_chunk_idx_start, k_chunk_idx_start, q, k, v, acc, l_prev, m_prev, ab, q_segment_ids, kv_segment_ids)
    if save_residuals:
        l, m = (v[..., 0] for v in aux[-2:])
        return (o, l, m)
    else:
        return o


def _flash_attention_dkv_kernel(
        q_chunk_idx_start_ref,
        k_chunk_idx_start_ref,
        q_tile_ref,
        k_tile_ref,
        v_tile_ref,
        ab_tile_ref,
        q_segment_ids_tile_ref,
        kv_segment_ids_tile_ref,
        l_tile_ref,
        m_tile_ref,
        do_tile_ref,
        di_tile_ref,
        dk_tile_ref,
        dv_tile_ref,
        dk_scratch_ref,
        dv_scratch_ref,
        *,
        sm_scale: float,
        causal: bool,
        mask_value: float,
        q_seq_len: int,
        block_q: int,
        block_k: int,
):
    _, _, block_q_major, _ = q_tile_ref.shape
    _, _, block_k_major, _ = k_tile_ref.shape

    q_seq_index = pl.program_id(axis=3)
    kv_seq_index = pl.program_id(axis=2)

    q_chunk_idx_start = q_chunk_idx_start_ref[0]
    k_chunk_idx_start = k_chunk_idx_start_ref[0]

    @pl.when(q_seq_index == 0)
    def start_new_sequence():
        dk_scratch_ref[:, :] = jnp.zeros(dk_scratch_ref.shape, dk_scratch_ref.dtype)
        dv_scratch_ref[:, :] = jnp.zeros(dv_scratch_ref.shape, dv_scratch_ref.dtype)

    def q_body(j, _):
        start_q = j * block_q

        def k_body(i, _):
            start_k = i * block_k
            k = pl.load(k_tile_ref, (0, 0, pl.ds(start_k, block_k), slice(None)))
            v = pl.load(v_tile_ref, (0, 0, pl.ds(start_k, block_k), slice(None)))
            q = pl.load(q_tile_ref, (0, 0, pl.ds(start_q, block_q), slice(None))
                        )  # [block_q, head_dim]
            l = pl.load(l_tile_ref, (0, 0, pl.ds(start_q, block_q), slice(None))
                        )  # [block_q, 128]
            m = pl.load(m_tile_ref, (0, 0, pl.ds(start_q, block_q), slice(None))
                        )  # [block_q, 128]
            do = pl.load(do_tile_ref, (0, 0, pl.ds(start_q, block_q), slice(None))
                         )  # [block_q, 128]
            di = pl.load(di_tile_ref, (0, 0, pl.ds(start_q, block_q), slice(None))
                         ).astype(jnp.float32)  # [block_q, 128]

            capped_logits = lax.dot_general(
                q, k, TRANS_B_DIM_NUMBERS, preferred_element_type=jnp.float32
            )  # [block_q_major, block_k]

            if ab_tile_ref is not None:
                ab = pl.load(
                    ab_tile_ref,
                    (
                        0,
                        0,
                        pl.dslice(j * block_q, block_q),
                        pl.dslice(i * block_k, block_k),
                    ),
                ).astype(jnp.float32)
                capped_logits += ab

            if sm_scale != 1.0:
                capped_logits *= sm_scale

            mask = None
            if q_segment_ids_tile_ref is not None:
                repeats, rem = divmod(block_k, NUM_LANES)
                if rem:
                    raise NotImplementedError(
                    )
                q_segment_ids = pl.load(
                    q_segment_ids_tile_ref, (0, pl.ds(start_q, block_q), slice(None))
                )  # [block_q, NUM_LANES].
                q_segment_ids = pltpu.repeat(
                    q_segment_ids, repeats, axis=1
                )  # [block_q, block_k].
                kv_segment_ids = pl.load(
                    kv_segment_ids_tile_ref, (slice(None), 0, pl.ds(start_k, block_k))
                )  # [1, block_k].
                mask = jnp.equal(q_segment_ids, kv_segment_ids).astype(jnp.bool_)

            if causal:
                mask_shape = (block_q, block_k)
                row_ids = jax.lax.broadcasted_iota(jnp.int32, mask_shape, 0)
                row_ids += (q_seq_index + q_chunk_idx_start) * block_q_major + start_q
                col_ids = jax.lax.broadcasted_iota(jnp.int32, mask_shape, 1)
                col_ids += (kv_seq_index + k_chunk_idx_start) * block_k_major + start_k
                causal_mask = col_ids <= row_ids
                mask = (
                    causal_mask if mask is None else jnp.logical_and(mask, causal_mask)
                )

            capped_logits = (
                capped_logits
                if mask is None
                else capped_logits + jnp.where(mask, 0.0, mask_value)
            )

            p = jnp.exp(
                capped_logits - pltpu.repeat(m, block_k // MIN_BLOCK_SIZE, axis=1)
            )
            p = p * pltpu.repeat(
                1 / l, block_k // MIN_BLOCK_SIZE, axis=1
            )  # [block_q_major, block_k_major]
            dv = lax.dot(p.T.astype(do.dtype), do, preferred_element_type=jnp.float32)
            pl.store(dv_scratch_ref, (pl.ds(start_k, block_k), slice(None)),
                     pl.load(dv_scratch_ref, (pl.ds(start_k, block_k), slice(None)))
                     + dv.astype(dv_scratch_ref.dtype))

            # di: [block_q, 128]
            # do: [block_q, head_dim]
            # v: [block_k_major, head_dim]
            dp = lax.dot_general(
                do, v, TRANS_B_DIM_NUMBERS, preferred_element_type=jnp.float32
            )
            ds = (dp - pltpu.repeat(di, block_k // MIN_BLOCK_SIZE, axis=1)) * p

            if sm_scale != 1.0:
                ds = ds * sm_scale

            # ds: [block_q_major, block_k_major]
            # q: [block_q_major, head_dim]
            dk = lax.dot(ds.T.astype(do.dtype), q, preferred_element_type=jnp.float32)
            pl.store(dk_scratch_ref, (pl.ds(start_k, block_k), slice(None)),
                     pl.load(dk_scratch_ref, (pl.ds(start_k, block_k), slice(None)))
                     + dk.astype(dk_scratch_ref.dtype))

        lax.fori_loop(0, block_k_major // block_k, k_body, None)

    if causal:
        should_run = below_or_on_diag(
            q_seq_index + q_chunk_idx_start, block_q_major, kv_seq_index + k_chunk_idx_start, block_k_major
        )
    else:
        should_run = True

    @pl.when(should_run)
    def run():
        lax.fori_loop(0, block_q_major // block_q, q_body, None)

    @pl.when(q_seq_index == q_seq_len // block_q_major - 1)
    def end_of_q_sequence():
        dv_tile_ref[0, 0, :, :] = dv_scratch_ref[...].astype(dv_tile_ref)
        dk_tile_ref[0, 0, :, :] = dk_scratch_ref[...].astype(dk_tile_ref)


def _flash_attention_bwd_dkv(
        q_chunk_idx_start,
        k_chunk_idx_start,
        q,
        k,
        v,
        ab,
        segment_ids,
        l,
        m,
        do,
        di,
        *,
        block_q_major: int | None,
        block_q: int | None,
        block_k_major: int | None,
        block_k: int | None,
        sm_scale: float,
        causal: bool = False,
        mask_value: float = DEFAULT_MASK_VALUE,
        debug: bool = False,
):
    batch_size, num_heads, q_seq_len, head_dim = q.shape
    _, _, kv_seq_len, _ = k.shape
    q_chunk_idx_start, k_chunk_idx_start = q_chunk_idx_start[None], k_chunk_idx_start[None]
    _verify_block("block_q_major_dkv", "q_seq_len", block_q_major, q_seq_len)
    _verify_block("block_q_dkv", "q_seq_len", block_q, q_seq_len)
    _verify_block("block_k_major_dkv", "kv_seq_len", block_k_major, kv_seq_len)
    _verify_block("block_k_dkv", "kv_seq_len", block_k, kv_seq_len)

    # Broadcast out scalar values
    m = jnp.broadcast_to(m[..., None], (*m.shape, MIN_BLOCK_SIZE))
    l = jnp.broadcast_to(l[..., None], (*l.shape, MIN_BLOCK_SIZE))
    # Preprocess contraction for bwd pass
    di = jnp.broadcast_to(di[..., None], (*di.shape, MIN_BLOCK_SIZE))

    # kv index needs to be before q index since q index is the contractng
    # dimension.
    grid = (
        batch_size,
        num_heads,
        kv_seq_len // block_k_major,
        q_seq_len // block_q_major,
    )

    def qo_index_map(batch_index, head_index, kv_seq_index, q_seq_index, q_idx_ref, k_idx_ref):
        if causal:
            # If the q block is skipped, stay at the 0th q block.
            next_q_index = lax.select(
                below_or_on_diag(
                    q_seq_index + q_idx_ref[0], block_q_major, kv_seq_index + k_idx_ref[0], block_k_major
                ),
                q_seq_index,
                0,
            )
        else:
            next_q_index = q_seq_index

        return (batch_index, head_index, next_q_index, 0)

    qo_spec = pl.BlockSpec(qo_index_map, (1, 1, block_q_major, head_dim))
    assert qo_spec.block_shape is not None
    assert q.ndim == len(qo_spec.block_shape)
    do_spec = qo_spec
    assert do.ndim == len(qo_spec.block_shape)

    def kv_index_map(batch_index, head_index, kv_seq_index, _, q_idx_ref, k_idx_ref):
        return (batch_index, head_index, kv_seq_index, 0)

    kv_spec = pl.BlockSpec(kv_index_map, (1, 1, block_k_major, head_dim))
    assert kv_spec.block_shape is not None
    assert k.ndim == len(kv_spec.block_shape)
    assert v.ndim == len(kv_spec.block_shape)

    def lm_index_map(batch_index, head_index, _, q_seq_index, q_idx_ref, k_idx_ref):
        return (batch_index, head_index, q_seq_index, 0)

    lm_spec = pl.BlockSpec(lm_index_map, (1, 1, block_q_major, MIN_BLOCK_SIZE))
    assert lm_spec.block_shape is not None
    assert l.ndim == len(lm_spec.block_shape)
    assert m.ndim == len(lm_spec.block_shape)

    di_spec = pl.BlockSpec(qo_index_map, (1, 1, block_q_major, MIN_BLOCK_SIZE))
    assert di_spec.block_shape is not None
    assert di.ndim == len(di_spec.block_shape)

    def ab_index_map(batch_index, head_index, kv_seq_index, q_seq_index, q_idx_ref, k_idx_ref):
        return (batch_index, kv_seq_index)

    dab_spec = (
        pl.BlockSpec(ab_index_map, (1, 1, block_q_major, block_k_major))
        if ab is not None
        else None
    )

    q_segment_ids_spec = kv_segment_ids_spec = None
    q_segment_ids = kv_segment_ids = None
    if segment_ids is not None:

        def q_segment_ids_index_map(
                batch_index, head_index, kv_seq_index, q_seq_index, q_idx_ref, k_idx_ref
        ):
            del head_index
            if causal:
                next_q_index = lax.select(
                    below_or_on_diag(
                        q_seq_index + q_idx_ref[0], block_q_major, kv_seq_index + k_idx_ref[0], block_k_major
                    ),
                    q_seq_index,
                    0,
                )
            else:
                next_q_index = q_seq_index
            return (batch_index, next_q_index, 0)

        def kv_segment_ids_index_map(batch_index, head_index, kv_seq_index, _, q_idx_ref, k_idx_ref):
            del head_index
            return (batch_index, 0, kv_seq_index)

        q_segment_ids_spec = pl.BlockSpec(
            q_segment_ids_index_map, (1, block_q_major, NUM_LANES)
        )
        kv_segment_ids_spec = pl.BlockSpec(
            kv_segment_ids_index_map, (1, NUM_SUBLANES, block_k_major)
        )

        q_segment_ids = jax.lax.broadcast_in_dim(
            segment_ids.q,
            (batch_size, q_seq_len, NUM_LANES),
            (
                0,
                1,
            ),
        )
        kv_segment_ids = jax.lax.broadcast_in_dim(
            segment_ids.kv,
            (batch_size, NUM_SUBLANES, kv_seq_len),
            (
                0,
                2,
            ),
        )

    in_specs = [
        qo_spec,
        kv_spec,
        kv_spec,
        dab_spec,
        q_segment_ids_spec,
        kv_segment_ids_spec,
        lm_spec,
        lm_spec,
        do_spec,
        di_spec,
    ]

    out_shapes = [
        jax.ShapeDtypeStruct((batch_size, num_heads, kv_seq_len, head_dim),
                             k.dtype),
        jax.ShapeDtypeStruct((batch_size, num_heads, kv_seq_len, head_dim),
                             v.dtype),
        jax.ShapeDtypeStruct((block_k_major, head_dim), jnp.float32),
        jax.ShapeDtypeStruct((block_k_major, head_dim), jnp.float32),
    ]

    def dkv_index_map(batch_index, head_index, kv_seq_index, _, q_idx_ref, k_idx_ref):
        return (batch_index, head_index, kv_seq_index, 0)

    dkv_spec = pl.BlockSpec(dkv_index_map, (1, 1, block_k_major, head_dim))
    out_specs = [
        dkv_spec, dkv_spec,
        pl.BlockSpec(lambda *_: (0, 0), (block_k_major, head_dim)),
        pl.BlockSpec(lambda *_: (0, 0), (block_k_major, head_dim)),
    ]

    kernel = functools.partial(
        _flash_attention_dkv_kernel,
        block_q=block_q,
        block_k=block_k,
        sm_scale=sm_scale,
        causal=causal,
        mask_value=mask_value,
        q_seq_len=q_seq_len,
    )
    name_scope = f"flash_mha_bwd_dkv_{block_q_major=}_{block_q=}_{block_k_major=}_{block_k=}"
    with jax.named_scope(name_scope):
        dk, dv, _, _ = pl.pallas_call(
            kernel,
            out_shape=out_shapes,
            grid_spec=pltpu.PrefetchScalarGridSpec(
                num_scalar_prefetch=2,
                in_specs=in_specs,
                out_specs=out_specs,
                grid=grid
            ),
            debug=debug,
            mosaic_params=dict(
                dimension_semantics=(
                    "parallel",
                    "parallel",
                    "parallel",
                    "arbitrary",
                )
            ),
        )(q_chunk_idx_start, k_chunk_idx_start, q, k, v, ab, q_segment_ids, kv_segment_ids, l, m, do, di)
        assert dk.shape == k.shape
        assert dv.shape == v.shape
    return dk, dv


def _flash_attention_dq_kernel(
        q_chunk_idx_start_ref,
        k_chunk_idx_start_ref,
        q_tile_ref,
        k_tile_ref,
        v_tile_ref,
        ab_tile_ref,
        q_segment_ids_tile_ref,
        kv_segment_ids_tile_ref,
        l_tile_ref,
        m_tile_ref,
        do_tile_ref,
        di_tile_ref,
        dq_tile_ref,
        dq_scratch_ref,
        ds_tile_ref,
        *,
        sm_scale: float,
        causal: bool,
        mask_value: float,
        kv_seq_len: int,
        block_k: int,
):
    _, _, block_k_major, _ = k_tile_ref.shape
    _, _, block_q_major, _ = q_tile_ref.shape

    kv_seq_index = pl.program_id(axis=3)
    q_seq_index = pl.program_id(axis=2)

    q_chunk_idx_start = q_chunk_idx_start_ref[0]
    k_chunk_idx_start = k_chunk_idx_start_ref[0]

    @pl.when(kv_seq_index == 0)
    def start_new_sequence():
        dq_scratch_ref[:, :] = jnp.zeros(dq_scratch_ref.shape, dq_scratch_ref.dtype)

    def body(i, _):
        k_slice = pl.ds(i * block_k, block_k)
        q = q_tile_ref[0, 0, :, :]
        k = pl.load(
            k_tile_ref, (0, 0, k_slice, slice(None)),
        )  # [block_k, head_dim]
        v = pl.load(
            v_tile_ref, (0, 0, k_slice, slice(None)),
        )  # [block_k, head_dim]
        l = l_tile_ref[0, 0, :, :]  # [block_q_major, 128]
        m = m_tile_ref[0, 0, :, :]  # [block_q_major, 128]
        do = do_tile_ref[0, 0, :, :]  # [block_q_major, head_dim]
        di = di_tile_ref[0, 0, :].astype(jnp.float32)  # [block_q_major, 128]

        capped_logits = jax.lax.dot_general(
            q, k, TRANS_B_DIM_NUMBERS, preferred_element_type=jnp.float32
        )

        if ab_tile_ref is not None:
            ab = pl.load(
                ab_tile_ref, (0, 0, pl.dslice(None), pl.dslice(i * block_k, block_k))
            ).astype(jnp.float32)
            capped_logits += ab

        if sm_scale != 1.0:
            capped_logits *= sm_scale

        mask = None
        if q_segment_ids_tile_ref is not None:
            repeats, rem = divmod(block_k, NUM_LANES)
            if rem:
                raise NotImplementedError(
                    f"kv block size must be a multiple of {NUM_LANES}"
                )
            q_segment_ids = pltpu.repeat(
                q_segment_ids_tile_ref[0], repeats, axis=1
            )  # [block_q, block_k].
            kv_segment_ids = pl.load(
                kv_segment_ids_tile_ref, (slice(None), 0, k_slice)
            )  # [1, block_k].
            mask = jnp.equal(q_segment_ids, kv_segment_ids).astype(jnp.bool_)

        if causal:
            mask_shape = (block_q_major, block_k)
            row_ids = jax.lax.broadcasted_iota(jnp.int32, mask_shape, 0)
            row_ids += (q_seq_index + q_chunk_idx_start) * block_q_major
            col_ids = jax.lax.broadcasted_iota(jnp.int32, mask_shape, 1)
            col_ids += (kv_seq_index + k_chunk_idx_start) * block_k_major + i * block_k
            causal_mask = col_ids <= row_ids
            mask = causal_mask if mask is None else jnp.logical_and(mask, causal_mask)
        capped_logits = (
            capped_logits
            if mask is None
            else capped_logits + jnp.where(mask, 0.0, mask_value)
        )

        p = jnp.exp(
            capped_logits - pltpu.repeat(m, block_k // MIN_BLOCK_SIZE, axis=1)
        )
        p = p * pltpu.repeat(
            1 / l, block_k // MIN_BLOCK_SIZE, axis=1
        )  # [block_q_major, block_k]

        # di: [block_q_major, 128]
        # do: [block_q_major, head_dim]
        # v: [block_k_major, head_dim]
        dp = jax.lax.dot_general(
            do,
            v,
            TRANS_B_DIM_NUMBERS,
            preferred_element_type=jnp.float32,
        )
        ds = (dp - pltpu.repeat(di, block_k // MIN_BLOCK_SIZE, axis=1)) * p
        # dp = jnp.dot(do, v.T)
        # ds = (dp - (dp * p).sum(axis=1)[:, None]) * p

        if sm_scale != 1.0:
            ds = ds * sm_scale

        if ds_tile_ref is not None:
            pl.store(
                ds_tile_ref,
                (0, 0, pl.dslice(None), pl.dslice(i * block_k, block_k)),
                ds.astype(ds_tile_ref.dtype),
            )

        # dp: [block_q_major, block_k]
        # k: [block_k, head_dim]
        dq_scratch_ref[:, :] += lax.dot(
            ds.astype(k.dtype),
            k,
            preferred_element_type=jnp.float32,
        ).astype(dq_scratch_ref.dtype)

    if causal:
        should_run = below_or_on_diag(
            q_seq_index + q_chunk_idx_start, block_q_major, kv_seq_index + k_chunk_idx_start, block_k_major
        )
        should_not_run = lax.select(should_run, False, True)
    else:
        should_run = True
        should_not_run = False  # type: ignore

    @pl.when(should_run)
    def run():
        lax.fori_loop(0, block_k_major // block_k, body, None)

    @pl.when(should_not_run)
    def zero_out_ds():
        if ds_tile_ref is not None:
            ds_tile_ref[...] = jnp.zeros_like(ds_tile_ref)

    @pl.when(kv_seq_index == kv_seq_len // block_k_major - 1)
    def end_of_kv_sequence():
        dq_tile_ref[0, 0, :, :] = dq_scratch_ref[...].astype(dq_tile_ref)
        dq_scratch_ref[...] = jnp.zeros_like(dq_scratch_ref)


def _flash_attention_bwd_dq(
        q_chunk_idx_start,
        k_chunk_idx_start,
        q,
        k,
        v,
        ab,
        segment_ids,
        l,
        m,
        do,
        di,
        *,
        block_q_major: int | None,
        block_k_major: int | None,
        block_k: int | None,
        sm_scale: float,
        causal: bool,
        mask_value: float,
        debug: bool,
):
    batch_size, num_heads, q_seq_len, head_dim = q.shape
    _, _, kv_seq_len, _ = k.shape
    q_chunk_idx_start, k_chunk_idx_start = q_chunk_idx_start[None], k_chunk_idx_start[None]
    _verify_block("block_q_dq", "q_seq_len", block_q_major, q_seq_len)
    _verify_block("block_k_major_dq", "kv_seq_len", block_k_major, kv_seq_len)
    _verify_block("block_k_dq", "block_k", block_k, kv_seq_len)

    # Broadcast out scalar values
    m = jnp.broadcast_to(m[..., None], (*m.shape, MIN_BLOCK_SIZE))
    l = jnp.broadcast_to(l[..., None], (*l.shape, MIN_BLOCK_SIZE))
    # Preprocess contraction for bwd pass
    di = jnp.broadcast_to(di[..., None], (*di.shape, block_k_major))

    grid = (
        batch_size,
        num_heads,
        q_seq_len // block_q_major,
        kv_seq_len // block_k_major,
    )

    def qo_index_map(batch_index, head_index, q_seq_index, _, q_idx_ref, k_idx_ref):
        return (batch_index, head_index, q_seq_index, 0)

    qo_spec = pl.BlockSpec(qo_index_map, (1, 1, block_q_major, head_dim))
    do_spec = qo_spec

    def kv_index_map(batch_index, head_index, q_seq_index, kv_seq_index, q_idx_ref, k_idx_ref):
        # if causal:
        #   # If the kv block is skipped, prefetch the next valid kv block, i.e. the
        #   # 0th one to be used for the next block_q rows.
        #   next_kv_index = lax.select(
        #       below_or_on_diag(
        #           q_seq_index + q_idx_ref[0], block_q_major, kv_seq_index + k_idx_ref[0], block_k_major
        #       ),
        #       kv_seq_index,
        #       0,
        #   )
        # else:
        next_kv_index = kv_seq_index
        return (batch_index, head_index, next_kv_index, 0)

    kv_spec = pl.BlockSpec(kv_index_map, (1, 1, block_k_major, head_dim))
    assert kv_spec.block_shape is not None
    assert k.ndim == len(kv_spec.block_shape)
    assert v.ndim == len(kv_spec.block_shape)

    def lm_index_map(batch_index, head_index, q_seq_index, _, q_idx_ref, k_idx_ref):
        return (batch_index, head_index, q_seq_index, 0)

    lm_spec = pl.BlockSpec(lm_index_map, (1, 1, block_q_major, MIN_BLOCK_SIZE))
    assert lm_spec.block_shape is not None
    assert l.ndim == len(lm_spec.block_shape)
    assert m.ndim == len(lm_spec.block_shape)

    di_spec = pl.BlockSpec(qo_index_map, (1, 1, block_q_major, MIN_BLOCK_SIZE))
    assert di_spec.block_shape is not None
    assert di.ndim == len(di_spec.block_shape)

    def ab_index_map(batch_index, head_index, q_seq_index, kv_seq_index, q_idx_ref, k_idx_ref):
        return (batch_index, head_index, q_seq_index, kv_seq_index)

    dab_spec = (
        pl.BlockSpec(ab_index_map, (1, 1, block_q_major, block_k_major))
        if ab is not None
        else None
    )

    q_segment_ids_spec = kv_segment_ids_spec = None
    q_segment_ids = kv_segment_ids = None
    if segment_ids is not None:
        assert False

        def q_segment_ids_index_map(batch_index, head_index, q_seq_index, _):
            del head_index
            return (batch_index, q_seq_index, 0)

        def kv_segment_ids_index_map(
                batch_index, head_index, q_seq_index, kv_seq_index
        ):
            del head_index
            if causal:
                # If the kv block is skipped, prefetch the next valid kv block, i.e. the
                # 0th one to be used for the next block_q rows.
                next_kv_index = lax.select(
                    below_or_on_diag(
                        q_seq_index, block_q_major, kv_seq_index, block_k_major
                    ),
                    kv_seq_index,
                    0,
                )
            else:
                next_kv_index = kv_seq_index
            return (batch_index, 0, next_kv_index)

        q_segment_ids_spec = pl.BlockSpec(
            q_segment_ids_index_map, (1, block_q_major, NUM_LANES)
        )
        kv_segment_ids_spec = pl.BlockSpec(
            kv_segment_ids_index_map, (1, NUM_SUBLANES, block_k_major)
        )

        q_segment_ids = jax.lax.broadcast_in_dim(
            segment_ids.q,
            (batch_size, q_seq_len, NUM_LANES),
            (
                0,
                1,
            ),
        )
        kv_segment_ids = jax.lax.broadcast_in_dim(
            segment_ids.kv,
            (batch_size, NUM_SUBLANES, kv_seq_len),
            (
                0,
                2,
            ),
        )

    in_specs = [
        qo_spec,
        kv_spec,
        kv_spec,
        dab_spec,
        q_segment_ids_spec,
        kv_segment_ids_spec,
        lm_spec,
        lm_spec,
        do_spec,
        di_spec,
    ]

    out_shapes = [
        jax.ShapeDtypeStruct(q.shape, q.dtype),
        jax.ShapeDtypeStruct((block_q_major, head_dim), jnp.float32),
        jax.ShapeDtypeStruct(ab.shape, ab.dtype) if ab is not None else None,
    ]
    dq_spec = pl.BlockSpec(qo_index_map, (1, 1, block_q_major, head_dim))
    out_specs = [
        dq_spec,
        pl.BlockSpec(lambda *_: (0, 0), (block_q_major, head_dim)),
        dab_spec,
    ]

    kernel = functools.partial(
        _flash_attention_dq_kernel,
        sm_scale=sm_scale,
        causal=causal,
        mask_value=mask_value,
        block_k=block_k,
        kv_seq_len=kv_seq_len,
    )
    name_scope = f"flash_mha_bwd_dq_{block_q_major=}_{block_k_major=}_{block_k=}"
    with jax.named_scope(name_scope):
        dq, _, ds = pl.pallas_call(
            kernel,
            out_shape=out_shapes,
            grid_spec=pltpu.PrefetchScalarGridSpec(
                num_scalar_prefetch=2,
                in_specs=in_specs,
                out_specs=out_specs,
                grid=grid
            ),
            debug=debug,
            mosaic_params=dict(
                dimension_semantics=(
                    "parallel",
                    "parallel",
                    "parallel",
                    "arbitrary",
                )
            ),
        )(q_chunk_idx_start, k_chunk_idx_start, q, k, v, ab, q_segment_ids, kv_segment_ids, l, m, do, di)

    # dab is just ds
    return dq, ds


# For autograd testing.
def mha_reference_no_custom_vjp(
        q,
        k,
        v,
        ab: jax.Array | None = None,
        segment_ids: SegmentIds | None = None,
        *,
        causal: bool = False,
        mask_value: float = DEFAULT_MASK_VALUE,
        sm_scale: float = 1.0,
        save_residuals: bool = False,
):
    logits = jnp.einsum("bhqc,bhkc->bhqk", q, k)
    if ab is not None:
        logits += ab
    if sm_scale != 1.0:
        logits *= sm_scale

    mask = None
    if segment_ids is not None:
        mask = segment_ids.q[:, :, None] == segment_ids.kv[:, None, :]
        mask = mask[:, None, :, :]

    if causal:
        _, _, q_seq_len, _ = q.shape
        _, _, kv_seq_len, _ = k.shape
        mask_shape = (q_seq_len, kv_seq_len)
        row_ids = jax.lax.broadcasted_iota(jnp.int32, mask_shape, 0)
        col_ids = jax.lax.broadcasted_iota(jnp.int32, mask_shape, 1)
        causal_mask = (col_ids <= row_ids)[None, None, :, :]
        mask = causal_mask if mask is None else jnp.logical_and(mask, causal_mask)

    logits = logits if mask is None else logits + jnp.where(mask, 0.0, mask_value)

    m = logits.max(axis=-1)
    unnormalized = jnp.exp(logits - m[..., None])
    l = unnormalized.sum(axis=-1)
    weights = unnormalized / l[..., None]
    out = jnp.einsum("bhqk,bhkc->bhqc", weights, v)
    if save_residuals:
        return out, l, m
    return out


@functools.partial(
    jax.jit, static_argnames=["causal", "mask_value", "sm_scale"]
)
@jax.default_matmul_precision("bfloat16")
def mha_reference(
        q,
        k,
        v,
        ab,
        segment_ids: SegmentIds | None = None,
        causal: bool = False,
        mask_value: float = DEFAULT_MASK_VALUE,
        sm_scale=1.0,
):
    return _mha_reference(
        q,
        k,
        v,
        ab,
        segment_ids,
        causal=causal,
        mask_value=mask_value,
        sm_scale=sm_scale,
        save_residuals=False,
    )


@functools.partial(jax.custom_vjp, nondiff_argnums=(5, 6, 7, 8))
def _mha_reference(
        q,
        k,
        v,
        ab,
        segment_ids: SegmentIds | None,
        causal: bool,
        mask_value: float,
        sm_scale: float,
        save_residuals: bool,
):
    return mha_reference_no_custom_vjp(
        q,
        k,
        v,
        ab,
        segment_ids,
        causal=causal,
        mask_value=mask_value,
        sm_scale=sm_scale,
        save_residuals=save_residuals,
    )


def _mha_reference_fwd(
        q,
        k,
        v,
        ab,
        segment_ids: SegmentIds | None,
        causal: bool,
        mask_value: float,
        sm_scale: float,
        save_residuals: bool,
):
    if save_residuals:
        raise NotImplementedError
    res = _mha_reference(
        q,
        k,
        v,
        ab,
        segment_ids,
        causal=causal,
        mask_value=mask_value,
        sm_scale=sm_scale,
        save_residuals=True,
    )
    assert isinstance(res, tuple)
    out, l, m = res
    return out, (q, k, v, ab, segment_ids, out, l, m)


@functools.partial(
    jax.jit,
    static_argnames=[
        "causal",
        "mask_value",
        "sm_scale",
    ],
)
def mha_reference_bwd(
        q,
        k,
        v,
        ab,
        segment_ids: SegmentIds | None,
        o,
        l,
        m,
        do,
        causal: bool = False,
        mask_value: float = DEFAULT_MASK_VALUE,
        sm_scale: float = 1.0,
):
    if sm_scale != 1.0:
        raise NotImplementedError

    logits = jnp.einsum(
        "bhqc,bhkc->bhqk",
        q.astype(jnp.float32),
        k.astype(jnp.float32),
    )
    if ab is not None:
        logits += ab

    mask = None
    if segment_ids is not None:
        mask = segment_ids.q[:, :, None] == segment_ids.kv[:, None, :]
        mask = mask[:, None, :, :]

    if causal:
        _, _, q_seq_len, _ = q.shape
        _, _, kv_seq_len, _ = k.shape
        mask_shape = (q_seq_len, kv_seq_len)
        row_ids = jax.lax.broadcasted_iota(jnp.int32, mask_shape, 0)
        col_ids = jax.lax.broadcasted_iota(jnp.int32, mask_shape, 1)
        causal_mask = (col_ids <= row_ids)[None, None, :, :]
        mask = causal_mask if mask is None else jnp.logical_and(mask, causal_mask)

    logits = logits if mask is None else logits + jnp.where(mask, 0.0, mask_value)

    unnormalized = jnp.exp(logits - m[..., None])
    p = unnormalized / l[..., None]
    dv = jnp.einsum("bhpt,bhpd->bhtd", p, do.astype(jnp.float32)).astype(v.dtype)

    dp = jnp.einsum(
        "bhpd,bhtd->bhpt", do.astype(jnp.float32), v.astype(jnp.float32)
    )

    di = jnp.sum(o.astype(jnp.float32) * do.astype(jnp.float32), axis=-1)[
        ..., None
    ]  # [batch_size, num_heads, q_seq_len]

    ds = (dp - di) * p
    dk = jnp.einsum("bhsd,bhst->bhtd", q.astype(jnp.float32), ds).astype(k.dtype)
    dq = jnp.einsum("bhst,bhtd->bhsd", ds, k.astype(jnp.float32)).astype(q.dtype)

    # dab is just ds
    dab = ds if ab is not None else None
    return dq, dk, dv, dab


def _mha_reference_bwd(
        causal: bool,
        mask_value: float,
        sm_scale: float,
        save_residuals: bool,
        residuals,
        do,
):
    del save_residuals
    q, k, v, ab, segment_ids, o, l, m = residuals
    dq, dk, dv, dab = mha_reference_bwd(
        q,
        k,
        v,
        ab,
        segment_ids,
        o,
        l,
        m,
        do,
        causal=causal,
        mask_value=mask_value,
        sm_scale=sm_scale,
    )
    return dq, dk, dv, dab, None


_mha_reference.defvjp(fwd=_mha_reference_fwd, bwd=_mha_reference_bwd)


def _verify_block(block_name, dim_name, block, dim, should_divide=True):
    if block > dim:
        raise ValueError(
            f"{block_name}={block} should be smaller or equal to {dim_name}={dim}"
        )
    if should_divide and dim % block != 0:
        raise ValueError(
            f"{dim_name}={dim} should be divisible by {block_name}={block}"
        )
